import numpy as np
import xarray as xr

import colorsys

import matplotlib
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import matplotlib.colors as colors
import matplotlib.patches as ptch
import matplotlib.collections as coll
import mpl_toolkits.axes_grid1 as axtk

from . import strat
from . import section

# plotting utilities


class VariableInfo(object):
    """Variable styling and information.

    This class holds information for a specific underlying data variable
    (e.g., ``eta`` or ``velocity``). Properties are used throughout
    DeltaMetrics for plotting.

    Examples
    --------

    You can override any of the parameters by passing a value during
    instantiation of the VariableInfo. For example, to set the colormap
    for a variable ``vegetation_density`` as a range of green shades we
    would create the VariableInfo as:

    .. doctest::

        >>> from deltametrics.plot import VariableInfo
        >>> veg = VariableInfo('vegetation_density',
        ...                    cmap='Greens')

        >>> veg.cmap
        <matplotlib.colors.LinearSegmentedColormap object at 0x...>

        >>> veg.cmap.N
        64

    which creates a 64 interval colormap from the matplotlib `Greens`
    colormap. If instead we wanted just three colors, we could specify the
    colormap manually:

    .. doctest::

        >>> veg3 = VariableInfo('vegetation_density',
        ...                     cmap=plt.cm.get_cmap('Greens', 3))
        >>> veg3.cmap.N
        3

    We can then set a more human-readable label for the variable:

    .. doctest::

        >>> veg3.label = 'vegetation density'

    """

    def __init__(self, name, **kwargs):
        """Initialize the VariableInfo.

        Parameters
        ----------
        name : :obj:`str`
            Name for variable to access from.

        cmap : :obj:`matplotlib.colors.Colormap`, :obj:`str`, optional
            Colormap to use to diplay the variable.

        label : :obj:`str`, optional
            How to display the variable name when plotting (i.e., a
            human-readable name). If not specified, the value of `name` is
            used.

        norm : :obj:`matplotlib.colors.Norm`, optional
            ???

        vmin : :obj:`float`, optional
            Limit the colormap or axes to this lower-bound value.

        vmax : :obj:`float`, optional
            Limit the colormap or axes to this upper-bound value.

        """
        if not type(name) is str:
            raise TypeError(
                'name argument must be type `str`, but was %s' % type(name))
        self._name = name

        self.cmap = kwargs.pop('cmap', cm.get_cmap('viridis', 64))
        self.label = kwargs.pop('label', None)
        self.norm = kwargs.pop('norm', None)
        self.vmin = kwargs.pop('vmin', None)
        self.vmax = kwargs.pop('vmax', None)

    @property
    def name(self):
        """Name for variable to access from.
        """
        return self._name

    @property
    def cmap(self):
        """Colormap to use to diplay the variable.
        """
        return self._cmap

    @cmap.setter
    def cmap(self, var):
        if type(var) is tuple:
            var, N = var
        else:
            N = 64
        if type(var) is str:
            self._cmap = cm.get_cmap(var, N)
        elif issubclass(type(var), colors.Colormap):
            self._cmap = var
        else:
            raise TypeError

    @property
    def label(self):
        """How to display the variable name when plotting.
        """
        return self._label

    @label.setter
    def label(self, var):
        if type(var) is str:
            self._label = var
        elif not var:
            self._label = self.name
        else:
            raise TypeError

    @property
    def norm(self):
        """???
        """
        return self._norm

    @norm.setter
    def norm(self, var):
        self._norm = var

    @property
    def vmin(self):
        """Limit the colormap or axes to this lower-bound value.
        """
        return self._vmin

    @vmin.setter
    def vmin(self, var):
        self._vmin = var

    @property
    def vmax(self):
        """Limit the colormap or axes to this upper-bound value.
        """
        return self._vmax

    @vmax.setter
    def vmax(self, var):
        self._vmax = var


class VariableSet(object):
    """A default set of properties for variables.

    This makes it easy to have consistent plots.

    This class defines a dictionary of VariableInfo objects, which store
    default information about that variable, for easy and consistent
    plotting.

    Additionally, you can create a VariableSet, and assign it to multiple
    :obj:`~deltametrics.cube.Cube` instances, so that each shares the same
    colormaps, etc.

    Examples
    --------

    Create a default variable set object.

    .. doctest::

        >>> dm.plot.VariableSet()
        <deltametrics.plot.VariableSet object at 0x...>

    """
    _after_init = False  # for "freezing" instance after init

    def __init__(self, override_dict=None):
        """Initialize the VariableSet.

        Initialize the set with default colormaps.

        .. note::
            It is expected that any attribute added to the `known_list` has a
            valid property defined in this class.

        .. note::
            We need an improved way to document each of the defaults. Want to
            display the colormaps, and the limits, the labels, etc. We want to
            be able to automatically label the image with low, high, labels,
            if there is aboundary norm, what happend to above and below
            values, etc. It's too verbose to do this in every file though,
            need a way to automate it. Either a custom Sphinx directive or mpl
            plot directives with function calls. There seems to be a bug that
            doesn't allow function calls to work for docstrings though, only
            primary documents. Another way could be to just loop all the
            ``known_list`` into a single fig at the end?

        Parameters
        ----------
        override_dict : :obj:`dict`, optional
            Dictionary defining variable-property sets. Dictionary key should
            set the  must be either a string (and then match defined colormaps
            in matplotlib or DeltaMetrics), a new matplotlib Colormap object,
            or an Mx3 numpy array that can be coerced into a linear colormap.
        """

        _added_list = ['net_to_gross']
        self.known_list = ['eta', 'stage', 'depth', 'discharge',
                           'velocity', 'sedflux', 'strata_sand_frac',
                           'sandfrac'] + \
                          ['x', 'y', 'time'] + _added_list

        self._variables = []
        for var in self.known_list:
            # set to defaults defined below (or None if no default)
            setattr(self, var, None)
            self._variables.append(var)

        if override_dict:  # loop override to set if given
            if not type(override_dict) is dict:
                raise TypeError('Invalid type for "override_dict".'
                                'Must be type dict, but was type: %s '
                                % type(override_dict))
            for var in override_dict:
                setattr(self, var, override_dict[var])
                self._variables.append(var)

        self._after_init = True

    @property
    def variables(self):
        """Variables known to the VariableSet."""
        return self._variables

    def __getitem__(self, var):
        """Get the attribute.

        Variable styling (i.e., the `VariableInfo` instances can be accessed by
        slicing the VariableSet with a string matching the VariableInfo `name`
        field. This enables accessing variables in evaluation, rather than
        explicit typing of variable names.

        Parameters
        ----------
        variable : :obj:`str`
            Which variable to get the `VariableInfo` for.

        .. note::

            If `variable` is not identified in the VariableSet, then a default
            instance of :obj:`VariableInfo` is instantiated with the variable
            name and returned.
        """
        if var in self._variables:
            return self.__getattribute__(var)
        else:
            return VariableInfo(var)

    def __setattr__(self, key, var):
        """Set, with check for types.

        This prevents setting non-VariableInfo attributes.
        """
        if not self._after_init:
            object.__setattr__(self, key, var)
        else:
            if type(var) is VariableInfo or var is None:
                object.__setattr__(self, key, var)
                # add it to the list of variables
                self._variables.append(key)
            else:
                raise TypeError(
                    'Can only set attributes of type VariableInfo.')

    @property
    def x(self):
        return self._x

    @x.setter
    def x(self, var):
        self._x = VariableInfo('x')

    @property
    def y(self):
        return self._y

    @y.setter
    def y(self, var):
        self._y = VariableInfo('y')

    @property
    def time(self):
        """Temporal history style.
        """
        return self._time

    @time.setter
    def time(self, var):
        if not var:
            self._time = VariableInfo('time')
        elif type(var) is VariableInfo:
            self._time = var
        else:
            raise TypeError

    @property
    def eta(self):
        """Bed elevation style.
        """
        return self._eta

    @eta.setter
    def eta(self, var):
        if not var:
            cmap = cm.get_cmap('cividis', 64)
            self._eta = VariableInfo('eta', cmap=cmap,
                                     label='bed elevation')
        elif type(var) is VariableInfo:
            self._eta = var
        else:
            raise TypeError

    @property
    def stage(self):
        """Flow stage style.
        """
        return self._stage

    @stage.setter
    def stage(self, var):
        if not var:
            self._stage = VariableInfo('stage')
        elif type(var) is VariableInfo:
            self._stage = var
        else:
            raise TypeError

    @property
    def depth(self):
        """Flow depth style.
        """
        return self._depth

    @depth.setter
    def depth(self, var):
        if not var:
            cmap = cm.get_cmap('Blues', 64)
            self._depth = VariableInfo('depth', cmap=cmap,
                                       vmin=0, label='flow depth')
        elif type(var) is VariableInfo:
            self._depth = var
        else:
            raise TypeError

    @property
    def discharge(self):
        """Flow discharge style.
        """
        return self._discharge

    @discharge.setter
    def discharge(self, var):
        if not var:
            cmap = cm.get_cmap('winter', 64)
            self._discharge = VariableInfo('discharge', cmap=cmap,
                                           label='flow discharge')
        elif type(var) is VariableInfo:
            self._discharge = var
        else:
            raise TypeError

    @property
    def velocity(self):
        """Flow velocity style.
        """
        return self._velocity

    @velocity.setter
    def velocity(self, var):
        if not var:
            cmap = cm.get_cmap('plasma', 64)
            self._velocity = VariableInfo('velocity', cmap=cmap,
                                          label='flow velocity')
        elif type(var) is VariableInfo:
            self._velocity = var
        else:
            raise TypeError

    @property
    def sedflux(self):
        """Flow sedflux style.
        """
        return self._sedflux

    @sedflux.setter
    def sedflux(self, var):
        if not var:
            cmap = cm.get_cmap('magma', 64)
            self._sedflux = VariableInfo('sedflux', cmap=cmap,
                                         label='sediment flux')
        elif type(var) is VariableInfo:
            self._sedflux = var
        else:
            raise TypeError

    @property
    def strata_sand_frac(self):
        """Sand fraction style.
        """
        return self._strata_sand_frac

    @strata_sand_frac.setter
    def strata_sand_frac(self, var):
        if not var:
            sandfrac = colors.ListedColormap(
                ['saddlebrown', 'sienna', 'goldenrod', 'gold'])
            sandfrac.set_under('saddlebrown')
            self._strata_sand_frac = VariableInfo('strata_sand_frac',
                                                  cmap=sandfrac,
                                                  norm=None, vmin=0,
                                                  label='sand fraction')
        elif type(var) is VariableInfo:
            self._strata_sand_frac = var
        else:
            raise TypeError

    @property
    def sandfrac(self):
        """Sand fraction style.
        """
        return self._sandfrac

    @sandfrac.setter
    def sandfrac(self, var):
        if not var:
            ends_str = ['saddlebrown',  'gold']  # define end points as strs
            endpts_colors = [matplotlib.colors.to_rgb(col) for col in ends_str]
            endpts_colors = np.column_stack(  # interpolate 64 between end pts
                [np.linspace(endpts_colors[0][i], endpts_colors[1][i], num=64)
                 for i in range(3)]  # for each column in RGB
                 )
            sand_frac = matplotlib.colors.ListedColormap(endpts_colors)
            sand_frac.set_under('saddlebrown')
            self._sandfrac = VariableInfo('sandfrac',
                                          cmap=sand_frac,
                                          norm=None, vmin=0,
                                          label='sand fraction')
        elif type(var) is VariableInfo:
            self._sandfrac = var
        else:
            raise TypeError

    @property
    def strata_depth(self):
        return self._strata_depth

    @strata_depth.setter
    def strata_depth(self, var):
        if not var:
            self._strata_depth = VariableInfo('strata_depth')
        elif type(var) is VariableInfo:
            self._strata_depth = var
        else:
            raise TypeError

    @property
    def net_to_gross(self):
        """Net-to-gross style.
        """
        return self._net_to_gross

    @net_to_gross.setter
    def net_to_gross(self, var):
        if not var:
            oranges = cm.get_cmap('Oranges', 64)
            greys = cm.get_cmap('Greys_r', 64)
            whiteblack = cm.get_cmap('Greys', 2)
            combined = np.vstack((greys(np.linspace(0.3, 0.6, 2)),
                                  oranges(np.linspace(0.2, 0.8, 6))))
            ntgcmap = colors.ListedColormap(combined, name='net_to_gross')
            ntgcmap.set_bad("white", alpha=0)
            self._net_to_gross = VariableInfo('net_to_gross',
                                              cmap=ntgcmap,
                                              label='net-to-gross')
        elif type(var) is VariableInfo:
            self.__net_to_gross = var
        else:
            raise TypeError


def cartographic_colormap(H_SL=0.0, h=4.5, n=1.0):
    """Colormap for an elevation map style.

    Parameters
    ----------
    H_SL : :obj:`float`, optional
        Sea level for the colormap. This is the break-point
        between blues and greens. Default value is `0.0`.

    h : :obj:`float`, optional
        Channel depth for the colormap. This is some characteristic *below
        sea-level* relief for the colormap to extend to through the range of
        blues. Default value is `4.5`.

    n : :obj:`float`, optional
        Surface topography relief for the colormap. This is some
        characteristic *above sea-level* relief for the colormap to extend to
        through the range of greens. Default value is `1.0`.

    Returns
    -------
    delta : :obj:`matplotib.colors.ListedColormap`
        The colormap object, which can then be used by other `matplotlib`
        plotting routines (see examples below).

    norm : :obj:`matplotib.colors.BoundaryNorm`
        The color normalization object, which can then be used by other
        `matplotlib` plotting routines (see examples below).

    Examples
    --------

    To display with default depth and relief parameters (left) and with adjust
    parameters to highlight depth variability (right):

    .. plot::
        :include-source:

        golfcube = dm.sample_data.golf()

        cmap0, norm0 = dm.plot.cartographic_colormap(H_SL=0)
        cmap1, norm1 = dm.plot.cartographic_colormap(H_SL=0, h=5, n=0.5)

        fig, ax = plt.subplots(1, 2, figsize=(10, 4))
        im0 = ax[0].imshow(golfcube['eta'][-1, ...], origin='lower',
                       cmap=cmap0, norm=norm0)
        cb0 = dm.plot.append_colorbar(im0, ax[0])
        im1 = ax[1].imshow(golfcube['eta'][-1, ...], origin='lower',
                       cmap=cmap1, norm=norm1)
        cb1 = dm.plot.append_colorbar(im1, ax[1])
        plt.show()
    """
    blues = matplotlib.cm.get_cmap('Blues_r', 64)
    greens = matplotlib.cm.get_cmap('YlGn_r', 64)
    combined = np.vstack((blues(np.linspace(0.1, 0.7, 5)),
                          greens(np.linspace(0.2, 0.8, 5))))
    delta = matplotlib.colors.ListedColormap(combined, name='delta')
    bounds = np.hstack(
        (np.linspace(H_SL-h, H_SL-(n/2), 5),
         np.linspace(H_SL, H_SL+n, 6)))
    norm = matplotlib.colors.BoundaryNorm(bounds, len(bounds)-1)
    return delta, norm


def aerial_colormap():
    """Colormap for a pesudorealistic looking aerial shot.

    .. warning::
        Not implemented.
    """

    raise NotImplementedError


def append_colorbar(ci, ax, size=2, pad=2, labelsize=9, **kwargs):
    """Append a colorbar, consistently placed.

    Adjusts some parameters of the parent axes as well.

    Parameters
    ----------
    ci : `matplotlib.pyplot.pcolormesh`, `matplotlib.pyplot.ImageAxes`
        The colored object generated via matplotlib that the colormap should
        be stolen from.

    ax : `matplotlib.Axes`
        The instance of axes to place the colorbar next to.

    size : :obj:`float`, optional
        Width (percent of parent axis width) of the colorbar; default is 2.

    pad : :obj:`float`, optional
        Padding between parent and colorbar axes. Default is 0.05.

    labelsize : :obj:`int`, optional
        Font size of label text. Default is 9pt.

    **kwargs
        Passed to `matplotlib.pyplot.colorbar`.

    Returns
    -------
    cb : `matplotlib.colorbar` instance.
        The colorbar instance created.
    """
    divider = axtk.axes_divider.make_axes_locatable(ax)
    cax = divider.append_axes("right", size=str(size)+"%", pad=str(pad)+"%")
    cb = plt.colorbar(ci, cax=cax, **kwargs)
    cb.ax.tick_params(labelsize=labelsize)
    ax.use_sticky_edges = False

    return cb

def style_axes_km(*args):
    """Style axes with kilometers, when initially set as meters.

    This function can be used two ways. Passing an `Axes` object as the first
    argument and optionally a string specifying `'x'`, `'y'`, or `'xy'`
    [default] will chnage the appearance of the `Axes` object. Alternatively,
    the function can be specified as the `ticker` function when setting a
    label formatter.

    Parameters
    ----------
    ax : :obj:`matplotlib.axes.Axes`
        Axes object to format

    which : :obj:`str`, optional
        Which axes to style as kilometers. Default is `xy`, but optionally
        pass `x` or `y` to only style one axis.

    x : :obj:`float`
        If using as a function to format labels, the tick mark location.

    Examples
    --------

    .. plot::
        :include-source:
        :context: reset

        golf = dm.sample_data.golf()

        fig, ax = plt.subplots(
            3, 1,
            gridspec_kw=dict(hspace=0.5))
        golf.quick_show('eta', ax=ax[0], ticks=True)

        golf.quick_show('eta', ax=ax[1], ticks=True)
        dm.plot.style_axes_km(ax[1])

        golf.quick_show('eta', axis=1, idx=10, ax=ax[2])
        ax[2].xaxis.set_major_formatter(dm.plot.style_axes_km)
        # OR use: dm.plot.style_axes_km(ax[2], 'x')

        plt.show()
    """
    if isinstance(args[0], matplotlib.axes.Axes):
        ax = args[0]
        if len(args) > 1:
            which = args[1]
        else:
            # default is to apply to both xy
            which = 'xy'
        # recursive calls to this func!
        if 'x' in which:
            ax.xaxis.set_major_formatter(style_axes_km)
        if 'y' in which:
            ax.yaxis.set_major_formatter(style_axes_km)

    else:
        v = args[0]
        return f'{v / 1000.:g}'


def get_display_arrays(VarInst, data=None):
    """Get arrays for display of Variables.

    Function takes as argument a `VariableInstance` from a `Section` or
    `Planform` and an optional :obj:`data` argument, which specifies the data
    type to return, and gives back arrays of 1) data, 2) display x-coordinates
    and 3) display y-coordinates.

    Parameters
    ----------
    VarInst : :obj:`~deltametrics.section.BaseSectionVariable` subclass
        The `Variable` instance to visualize. May be any subclass of
        :obj:`~deltametrics.section.BaseSectionVariable` or
        :obj:`~deltametrics.plan.BasePlanformVariable`.

    data : :obj:`str`, optional
        The type of data to visualize. Supported options are `'spacetime'`,
        `'preserved'`, and `'stratigraphy'`. Default is to display full
        spacetime plot for variable generated from a `DataCube`, and
        stratigraphy for a `StratigraphyCube` variable. Note that variables
        sourced from a `StratigraphyCube` do not support the `spacetime` or
        `preserved` options, and a variable from `DataCube` will only support
        `stratigraphy` if the Cube has computed "quick" stratigraphy.

    Returns
    -------
    data, X, Y : :obj:`ndarray`
        Three matching-size `ndarray` representing the 1) data, 2) display
        x-coordinates and 3) display y-coordinates.
    """
    # # #  SectionVariables  # # #
    if VarInst.slicetype == 'data_section':
        # #  DataSection  # #
        data = data or 'spacetime'
        if data in VarInst.strat._spacetime_names:
            _S, _Z = np.meshgrid(VarInst['s'], VarInst[VarInst.dims[0]])
            return VarInst.values, _S, _Z
        elif data in VarInst.strat._preserved_names:
            VarInst.strat._check_knows_spacetime()
            _S, _Z = np.meshgrid(VarInst['s'], VarInst[VarInst.dims[0]])
            return VarInst.strat.as_preserved(), _S, _Z
        elif data in VarInst.strat._stratigraphy_names:
            _sp = VarInst.strat.as_stratigraphy()
            _den = _sp.toarray()  # .view(section.DataSectionVariable)
            _arr_Y = VarInst.strat.strat_attr['psvd_flld'][:_sp.shape[0], ...]
            _arr_X = np.tile(VarInst['s'], (_sp.shape[0], 1))
            return _den[1:, 1:], _arr_X, _arr_Y
        else:
            raise ValueError('Bad data argument: %s' % str(data))

    elif VarInst.slicetype == 'stratigraphy_section':
        # #  StratigraphySection  # #
        data = data or 'stratigraphy'
        if data in VarInst.strat._spacetime_names:
            VarInst.strat._check_knows_spacetime()  # always False
        elif data in VarInst.strat._preserved_names:
            VarInst.strat._check_knows_spacetime()  # always False
        elif data in VarInst.strat._stratigraphy_names:
            _S, _Z = np.meshgrid(VarInst['s'], VarInst[VarInst.dims[0]])
            return VarInst, _S, _Z
        else:
            raise ValueError('Bad data argument: %s' % str(data))

    # # #  PlanformVariables  # # #
    elif False:  # issubclass(type(VarInst), plan.BasePlanformVariable):
        raise NotImplementedError
    else:
        raise TypeError('Invaid "VarInst" type: %s' % type(VarInst))


def get_display_lines(VarInst, data=None):
    """Get lines for display of Variables.

    Function takes as argument a `VariableInstance` from a `Section` or
    `Planform` and an optional :obj:`data` argument, which specifies the data
    type to return, and gives back data and line segments for display.

    Parameters
    ----------
    VarInst : :obj:`~deltametrics.section.BaseSectionVariable` subclass
        The `Variable` instance to visualize. May be any subclass of
        :obj:`~deltametrics.section.BaseSectionVariable` or
        :obj:`~deltametrics.plan.BasePlanformVariable`.

    data : :obj:`str`, optional
        The type of data to visualize. Supported options are `'spacetime'`,
        `'preserved'`, and `'stratigraphy'`. Default is to display full
        spacetime plot for variable generated from a `DataCube`.  and
        stratigraphy for a `StratigraphyCube` variable. Variables from
        `DataCube` will only support `stratigraphy` if the Cube has computed
        "quick" stratigraphy.

    .. note::
        Not currently implemented for variables from the `StratigraphyCube`.

    Returns
    -------
    vals, segments : :obj:`ndarray`
        An off-by-one sized array of data values (`vals`) to color the line
        segments (`segments`). The segments are organized along the 0th
        dimension of the array.
    """
    # # #  SectionVariables  # # #
    if VarInst.slicetype == 'data_section':
        # #  DataSection  # #
        def _reshape_long(X):
            # util for reshaping s- and z-values appropriately
            return np.vstack((X[:, :-1].flatten(),
                              X[:, 1:].flatten())).T.reshape(-1, 2, 1)
        data = data or 'spacetime'
        _S, _Z = np.meshgrid(VarInst['s'], VarInst[VarInst.dims[0]])
        if data in VarInst.strat._spacetime_names:
            z = _reshape_long(_Z)
            vals = VarInst[:, :-1]
        elif data in VarInst.strat._preserved_names:
            z = _reshape_long(_Z)
            vals = VarInst.strat.as_preserved()[:, :-1]
        elif data in VarInst.strat._stratigraphy_names:
            VarInst.strat._check_knows_stratigraphy()  # need to check explicitly
            z = _reshape_long(np.copy(VarInst.strat.strat_attr['strata']))
            vals = VarInst[:, :-1]
        else:
            raise ValueError('Bad data argument: %s' % str(data))
        s = _reshape_long(_S)
        segments = np.concatenate([s, z], axis=2)
        if data in VarInst.strat._stratigraphy_names:
            # flip = draw late to early
            vals = np.fliplr(np.flipud(vals))
            segments = np.flipud(segments)
        return np.array(vals), segments

    elif VarInst.slicetype == 'stratigraphy_section':
        # #  StratigraphySection  # #
        data = data or 'stratigraphy'
        if data in VarInst.strat._spacetime_names:
            VarInst.strat._check_knows_spacetime()  # always False
        elif data in VarInst.strat._preserved_names:
            VarInst.strat._check_knows_spacetime()  # always False
        elif data in VarInst.strat._stratigraphy_names:
            raise NotImplementedError  # not sure best implementation
        else:
            raise ValueError('Bad data argument: %s' % str(data))

    # # #  PlanformVariables  # # #
    elif False:  # issubclass(type(VarInst), plan.BasePlanformVariable):
        raise NotImplementedError
    else:
        raise TypeError('Invaid "VarInst" type: %s' % type(VarInst))


def get_display_limits(VarInst, data=None, factor=1.5):
    """Get limits to resize the display of Variables.

    Function takes as argument a `VariableInstance` from a `Section` or
    `Planform` and an optional :obj:`data` argument, which specifies how to
    determine the limits to return.

    Parameters
    ----------
    VarInst : :obj:`~deltametrics.section.BaseSectionVariable` subclass
        The `Variable` instance to visualize. May be any subclass of
        :obj:`~deltametrics.section.BaseSectionVariable` or
        :obj:`~deltametrics.plan.BasePlanformVariable`.

    data : :obj:`str`, optional
        The type of data to compute limits for. Typically this will be the
        same value used with either :obj:`get_display_arrays` or
        :obj:`get_display_lines`. Supported options are `'spacetime'`,
        `'preserved'`, and `'stratigraphy'`.

    factor : :obj:`float`, optional
        Factor to extend vertical limits upward for stratigraphic sections.

    Returns
    -------
    xmin, xmax, ymin, ymax : :obj:`float`
        Values to use as limits on a plot. Use with, for example,
        ``ax.set_xlim((xmin, xmax))``.
    """
    # # #  SectionVariables  # # #
    if VarInst.slicetype == 'data_section':
        # #  DataSection  # #
        data = data or 'spacetime'
        _S, _Z = np.meshgrid(VarInst['s'], VarInst[VarInst.dims[0]])
        if data in VarInst.strat._spacetime_names:
            return np.min(_S), np.max(_S), \
                np.min(_Z), np.max(_Z)
        elif data in VarInst.strat._preserved_names:
            VarInst.strat._check_knows_stratigraphy()  # need to check explicitly
            return np.min(_S), np.max(_S), \
                np.min(_Z), np.max(_Z)
        elif data in VarInst.strat._stratigraphy_names:
            VarInst.strat._check_knows_stratigraphy()  # need to check explicitly
            _strata = np.copy(VarInst.strat.strat_attr['strata'])
            return np.min(_S), np.max(_S), \
                np.min(_strata), np.max(_strata) * factor
        else:
            raise ValueError('Bad data argument: %s' % str(data))

    elif VarInst.slicetype == 'stratigraphy_section':
        # #  StratigraphySection  # #
        data = data or 'stratigraphy'
        _S, _Z = np.meshgrid(VarInst['s'], VarInst[VarInst.dims[0]])
        if data in VarInst.strat._spacetime_names:
            VarInst.strat._check_knows_spacetime()  # always False
        elif data in VarInst.strat._preserved_names:
            VarInst.strat._check_knows_spacetime()  # always False
        elif data in VarInst.strat._stratigraphy_names:
            return np.min(_S), np.max(_S), \
                np.min(_Z), np.max(_Z) * factor
        else:
            raise ValueError('Bad data argument: %s' % str(data))

    # # #  PlanformVariables  # # #
    elif False:  # issubclass(type(VarInst), plan.BasePlanformVariable):
        raise NotImplementedError
    else:
        raise TypeError('Invaid "VarInst" type: %s' % type(VarInst))


def _fill_steps(where, x=1, y=1, y0=0, **kwargs):
    """Fill rectangles where the boolean indicates ``True``.

    Creates an :obj:`x` by :obj:`y` :obj:`matplotlib.patches.Rectangle` at
    each index where the index `i` in boolean :obj:`where` indicates, with the
    lower left of the patch at (`i`, :obj:`y0`).

    This utility function is utilized internally. Most often, it is used to
    highlight where time has been preserved (or eliminated) in a timeseries of
    data. For example, see
    :obj:`~deltametrics.plot.show_one_dimensional_trajectory_to_strata`.

    Parameters
    ----------
    where : :obj:`ndarray`
        Boolean `numpy` `ndarray` indicating which locations to create a patch
        at.

    x : :obj:`float`
        The x-direction width of the `Rectangles`.

    y : :obj:`float`
        The y-direction height of the `Rectangles`.

    y0 : :obj:`float`
        The y-direction origin (i.e., lower-left y-value) of the `Rectangles`.

    **kwargs
        Additional `matplotlib` keyword arguments passed to the
        `Rectangle` instantiation. This is where you can set most
        `matplotlib` `**kwargs` (e.g., `color`, `edgecolor`).

    Returns
    -------
    pc : :obj:`matplotlib.patches.PatchCollection`
        Collection of `Rectangle` `Patch` objects.
    """
    pl = []
    for i, pp in enumerate(np.argwhere(where[1:]).flatten()):
        _r = ptch.Rectangle((pp, y0), x, y, **kwargs)
        pl.append(_r)
    return coll.PatchCollection(pl, match_original=True)


def show_one_dimensional_trajectory_to_strata(e, sigma_dist=None,
                                              dz=None, z=None,
                                              nz=None, ax=None,
                                              show_strata=True,
                                              label_strata=False):
    """1d elevation to stratigraphy.

    This function creates and displays a one-dimensional elevation timeseries
    as a trajectory of elevations, resultant stratigraphy, and time preserved
    in "boxy" stratigraphy. The function is helpful for description of
    algorithms to compute stratigraphy, and for debugging and checking outputs
    from computations.

    For example, we can quickly visualize the processing of a 1D timeseries of
    bed elevations into boxy stratigraphy with this routine.

    .. plot:: guides/userguide_1d_example.py
        :include-source:

    The orange line depicts the resultant stratigraphy, with all
    bed-elevations above this line cut from the stratigraphy by the
    stratigraphic filter.  The column on the right records which
    time-interval is recorded in the stratigraphy at each elevation.

    Parameters
    ----------
    e : :obj:`ndarray`
        Elevation data as a 1D array.

    sigma_dist : :obj:`ndarray`, :obj:`float`, :obj:`int`, optional
        Optional subsidence distance argument used to adjust the elevation
        data to account for subsidence when computing stratigraphy. See
        :obj:`_adjust_elevation_by_subsidence` for a complete description.

    z : :obj:`ndarray`, optional
        Vertical coordinates for stratigraphy, in meters. Optional, and
        mutually exclusive with :obj:`dz` and :obj:`nz`,
        see :obj:`_determine_strat_coordinates` for complete description.

    dz : :obj:`float`, optional
        Vertical resolution of stratigraphy, in meters. Optional, and mutually
        exclusive with :obj:`z` and :obj:`nz`,
        see :obj:`_determine_strat_coordinates` for complete description.

    nz : :obj:`int`, optional
        Number of intervals for vertical coordinates of stratigraphy.
        Optional, and mutually exclusive with :obj:`z` and :obj:`dz`,
        see :obj:`_determine_strat_coordinates` for complete description.

    ax : :obj:`matplotlib.pyplot.axes`, optional
        Axes to plot into. A figure and axes is created, if not given.

    show_strata : :obj:`bool`, optional
        Whether to plot the resultant stratigraphy as a strip log on a second
        axis on the right side. This axis included numbers indicating the
        timestep for each voxel of preserved stratigraphy.

    Returns
    -------
    """
    # reprocess shape to be 1d array, if needed.
    if e.ndim > 1:
        if e.shape[1:] == (1, 1):
            e = e.squeeze()
        else:
            raise ValueError('Elevation data "e" must be one-dimensional.')
    t = np.arange(e.shape[0])  # x-axis time array
    t3 = np.expand_dims(t, axis=(1, 2))  # 3d time, for slicing
    e_in = e.copy()

    if sigma_dist is not None:
        # adjust elevations by subsidence rate
        e = strat._adjust_elevation_by_subsidence(e_in, sigma_dist)
    s, p = strat._compute_elevation_to_preservation(e)  # strat, preservation
    z = strat._determine_strat_coordinates(e, dz=dz, z=z, nz=nz)  # vert coordinates
    sc, dc = strat._compute_preservation_to_cube(s, z)
    lst = np.argmin(s < s[-1])  # last elevation

    c = np.full_like(z, np.nan)
    c[sc[:, 0]] = t[dc[:, 0]]
    cp = np.tile(c, (2, 1)).T

    # make the plots
    if not ax:
        fig, ax = plt.subplots()
    # timeseries plot
    pt = np.zeros_like(t)  # for psvd timesteps background
    pt[np.union1d(p.nonzero()[0], np.array(
        strat._compute_preservation_to_time_intervals(p).nonzero()[0]))] = 1
    ax.add_collection(_fill_steps(p, x=1, y=np.max(e_in) - np.min(e),
                                  y0=np.min(e), facecolor='0.8'))
    ax.add_patch(ptch.Rectangle((0, 0), 0, 0, facecolor='0.8',
                                label='psvd timesteps'))  # add for lgnd
    ax.hlines(s[p], 0, e.shape[0], linestyles='dashed', colors='0.7')
    ax.axvline(lst, c='k')
    ax.step(t, e_in, where='post', label='elevation')
    ax.step(t, s, linestyle='--', where='post', label='stratigraphy')
    ax.plot(t[p], s[p], color='0.5', marker='o',
            ls='none', label='psvd time')
    if len(t) < 100:
        ax.xaxis.set_minor_locator(matplotlib.ticker.MultipleLocator(1))
    ax.grid(which='both', axis='x')

    if show_strata:
        # boxy strata plot
        divider = axtk.make_axes_locatable(ax)
        ax_s = divider.append_axes("right", 0.5, pad=0.1, sharey=ax)
        ax_s.yaxis.tick_right()
        ax_s.xaxis.set_visible(False)
        __x, __y = np.meshgrid(np.array([0, 1]), z)
        _colmap = plt.cm.get_cmap('viridis', e.shape[0])
        ax_s.pcolormesh(__x, __y, cp,
                        cmap=_colmap, vmin=0, vmax=e.shape[0],
                        shading='auto')
        ax_s.hlines(e[p], 0, 1, linestyles='dashed', colors='gray')
        _cstr = [str(int(cc)) if np.isfinite(cc) else 'nan' for cc in c.flatten()]
        ax_s.set_xlim(0, 1)
        if label_strata:
            for i, __cstr in enumerate(_cstr):
                ax_s.text(0.3, z[i], str(__cstr), fontsize=8)

    # adjust and add legend
    if np.any(e < 0):
        ax.set_ylim(np.min(e) * 1.2, np.maximum(0, np.max(e_in) * 1.2))
    else:
        ax.set_ylim(np.min(e) * 0.8, np.max(e_in) * 1.2)
    ax.legend()


def _scale_lightness(rgb, scale_l):
    """Utility to scale the lightness of some color.

    Make a color perceptually lighter or darker. Adapted from
    https://stackoverflow.com/a/60562502/4038393.

    Parameters
    ----------
    rgb : :obj:`tuple`
        A three element tuple of the RGB values for the color.

    scale_l : :obj:`float`
        The scale factor, relative to a value of `1`. A value of 1 performs no
        scaling, and values less than 1 darken the color, whereas values
        greater than 1 brighten the color.

    Returns
    -------
    scaled : :obj:`tuple`
        Scaled color RGB tuple.

    Example
    -------

    .. plot::

        fig, ax = plt.subplots(figsize=(5, 2))

        # initial color red
        red = (1.0, 0.0, 0.0)
        ax.plot(-1, 1, 'o', color=red)

        # scale from 1 to 0.05
        scales = np.arange(1, 0, -0.05)

        # loop through scales and plot
        for s, scale in enumerate(scales):
            darker_red = dm.plot._scale_lightness(red, scale)
            ax.plot(s, scale, 'o', color=darker_red)

        plt.show()
    """
    # https://stackoverflow.com/a/60562502/4038393
    # convert rgb to hls
    h, l, s = colorsys.rgb_to_hls(*rgb)
    # manipulate h, l, s values and return as rgb
    return colorsys.hls_to_rgb(h, min(1, l * np.abs(scale_l)), s=s)


def show_histograms(*args, sets=None, ax=None, **kwargs):
    """Show multiple histograms, including as sets.

    Parameters
    ----------
    *args : :obj:`tuple`
        Any number of comma separated tuples, where each tuple is a set of
        `(counts, bins)`, for example, as an output from `np.histogram()`.

    sets : :obj:`list`, optional
        A list or numpy array indicating the set each pdf belongs to. For
        example, [0, 0, 1, 1, 2] incidates the first two `*args` are from the
        first set, the third and fourth belong to a second set, and the fifth
        argument belongs to a third set. Length of `sets` must match the
        number of comma separated `*args`. If not supplied, all histograms are
        colored differently (up to 10).

    ax : :obj:`matplotlib.pyplot.axes`, optional
        Axes to plot into. A figure and axes is created, if not given.

    **kwargs
        Additional `matplotlib` keyword arguments passed to the
        `bar` plotting routine. In current implementation, cannot use
        arguments `width`, `edgecolor`, or `facecolor`.

    Returns
    -------

    Examples
    --------

    .. plot::
        :include-source:

        locs = [0.25, 1, 0.5, 4, 2]
        scales = [0.1, 0.25, 0.4, 0.5, 0.1]
        bins = np.linspace(0, 6, num=40)

        hist_bin_sets = [np.histogram(np.random.normal(l, s, size=500), bins=bins, density=True) for l, s in zip(locs, scales)]

        fig, ax = plt.subplots()
        dm.plot.show_histograms(*hist_bin_sets, sets=[0, 1, 0, 1, 2], ax=ax)
        ax.set_xlim((0, 6))
        ax.set_ylabel('density')
        plt.show()
    """
    if not ax:
        fig, ax = plt.subplots()

    if (sets is None):
        n_sets = len(args)
        sets = np.arange(n_sets)
    else:
        n_sets = len(np.unique(sets))
        sets = np.array(sets)

    if len(sets) != len(args):
        raise ValueError(
            'Number of histogram tuples must match length of `sets` list.')

    for i in range(n_sets):
        CN = 'C%d' % (i)
        match = np.where((sets == i))[0]
        scales = np.linspace(0.8, 1.2, num=len(match))
        CNs = [_scale_lightness(colors.to_rgb(CN), sc) for s, sc in enumerate(scales)]
        for n in range(len(match)):
            hist, bins = args[match[n]]
            bin_width = (bins[1:] - bins[:-1])
            bin_cent = bins[:-1] + (bin_width/2)
            ax.bar(bin_cent, hist, width=bin_width,
                   edgecolor=CNs[n], facecolor=CNs[n], **kwargs)


def aerial_view(elevation_data, datum=0, ax=None, ticks=False,
                colorbar_kw={}, return_im=False, **kwargs):
    """Show an aerial plot of an elevation dataset.

    See also: implementation wrapper for a cube.

    Parameters
    ----------

    elevation_data
        2D array of elevations.

    datum
        Sea level reference, default value is 0.

    ax : :obj:`~matplotlib.pyplot.Axes` object, optional
        A `matplotlib` `Axes` object to plot the section. Optional; if not
        provided, a call is made to ``plt.gca()`` to get the current (or
        create a new) `Axes` object.

    ticks
        Whether to show ticks. Default is false, no tick labels.

    colorbar_kw
        Dictionary of keyword args passed to :func:`append_colorbar`.

    return_im : bool, optional
        Returns the ``plt.imshow()`` image object if True. Default is False.

    **kwargs
        Optionally, arguments accepted by :func:`cartographic_colormap`, or
        `imshow`.

    Returns
    -------
    colorbar
        Colorbar instance appended to the axes.

    im : :obj:`~matplotlib.image.AxesImage` object, optional
        Optional return of the image object itself if ``return_im`` is True


    Examples
    --------
    .. plot::
        :include-source:

        golfcube = dm.sample_data.golf()
        elevation_data = golfcube['eta'][-1, :, :]

        fig, ax = plt.subplots()
        dm.plot.aerial_view(elevation_data, ax=ax)
        plt.show()

    """
    if not ax:
        fig, ax = plt.subplots()

    # process to make a cmap
    h = kwargs.pop('h', 3)
    n = kwargs.pop('n', 1)
    carto_cm, carto_norm = cartographic_colormap(H_SL=0, h=h, n=n)

    # get the extent to plot
    if isinstance(elevation_data, xr.core.dataarray.DataArray):
        d0, d1 = elevation_data.dims
        d0_arr, d1_arr = elevation_data[d0], elevation_data[d1]
        _extent = [d1_arr[0],                  # dim1, 0
                   d1_arr[-1] + d1_arr[1],     # dim1, end + dx
                   d0_arr[-1] + d0_arr[1],     # dim0, end + dx
                   d0_arr[0]]                  # dim0, 0
    else:
        _extent = [0, elevation_data.shape[1],
                   elevation_data.shape[0], 0]

    # plot the data
    im = ax.imshow(
        elevation_data - datum,
        cmap=carto_cm, norm=carto_norm,
        extent=_extent, **kwargs)

    cb = append_colorbar(im, ax, **colorbar_kw)
    if not ticks:
        ax.set_xticks([], minor=[])
        ax.set_yticks([], minor=[])

    if return_im is True:
        return cb, im
    else:
        return cb


def overlay_sparse_array(sparse_array, ax=None, cmap='Reds',
                         alpha_clip=(None, 90), clip_type='percentile'):
    """Convenient plotting method to overlay a sparse 2D array on an image.

    Should only be used with data arrays that are sparse: i.e., where many
    elements take the value 0 (or very near 0).

    Original implementation was borrowed from the `dorado` project's
    implementation of `show_nourishment_area`.

    Parameters
    ----------
    sparse_array
        2d array to plot.

    ax : :obj:`~matplotlib.pyplot.Axes` object, optional
        A `matplotlib` `Axes` object to plot the section. Optional; if not
        provided, a call is made to ``plt.gca()`` to get the current (or
        create a new) `Axes` object.

    cmap
        Maplotlib colormap to use. String or colormap object.

    alpha_clip
        Two element tuple, specifying the *percentiles* of the data in
        `sparse_array` to clip for display. First element specifies the lower
        bound to clip, second element is the upper bound. Either or both
        elements can be `None` to indicate no clipping. Default is ``(None,
        90)``.

    clip_type
        String specifying how `alpha_clip` should be interpreted. Accepted
        values are `'percentile'` (default) and `'value'`. If `'percentile'`,
        the data in `sparse_array` are clipped based on the density of the
        data at the specified percentiles; note values should be in range
        [0, 100). If  `'value'`, the data in `sparse_array` are clipped
        directly based on the values in `alpha_clip`.

    Returns
    -------
    image
        image object returned from `imshow`.

    Examples
    --------
    Here, we use the discharge field from the model as an example
    of sparse data.

    .. plot::
        :include-source:
        :context: reset

        golfcube = dm.sample_data.golf()
        elevation_data = golfcube['eta'][-1, :, :]
        sparse_data = golfcube['discharge'][-1, ...]

        fig, ax = plt.subplots(1, 3, figsize=(8, 3))
        for axi in ax.ravel():
            dm.plot.aerial_view(elevation_data, ax=axi)

        dm.plot.overlay_sparse_array(
            sparse_data, ax=ax[0])  # default clip is (None, 90)
        dm.plot.overlay_sparse_array(
            sparse_data, alpha_clip=(None, None), ax=ax[1])
        dm.plot.overlay_sparse_array(
            sparse_data, alpha_clip=(70, 90), ax=ax[2])

        plt.tight_layout()
        plt.show()

    .. plot::
        :include-source:
        :context: close-figs

        fig, ax = plt.subplots(1, 3, figsize=(8, 3))
        for axi in ax.ravel():
            dm.plot.aerial_view(elevation_data, ax=axi)

        dm.plot.overlay_sparse_array(
            sparse_data, ax=ax[0],
            clip_type='value')  # default clip is (None, 90)
        dm.plot.overlay_sparse_array(
            sparse_data, ax=ax[1],
            alpha_clip=(None, 0.2), clip_type='value')
        dm.plot.overlay_sparse_array(
            sparse_data, ax=ax[2],
            alpha_clip=(0.4, 0.6), clip_type='value')

        plt.tight_layout()
        plt.show()

    """
    if not ax:
        fig, ax = plt.subplots()

    # check this is a tuple or list
    if isinstance(alpha_clip, tuple) or isinstance(alpha_clip, list):
        if len(alpha_clip) != 2:
            raise ValueError(
                '`alpha_clip` must be tuple or list of length 2.')
    else:  # if it is a tuple, check the length
        raise TypeError(
            '`alpha_clip` must be type `tuple`, '
            'but was type {0}.'.format(type(alpha_clip)))

    # check the clip_type flag
    clip_type_allow = ['percentile', 'value']
    if clip_type not in clip_type_allow:
        raise ValueError(
            'Bad value given for `clip_type` argument. Input argument must '
            'be one of `{0}`, but was `{1}`'.format(
                clip_type_allow, clip_type))

    # pull the cmap out
    if isinstance(cmap, str):
        cmap = plt.cm.get_cmap(cmap)
    else:
        cmap = cmap

    # get the extent to plot
    if isinstance(sparse_array, xr.core.dataarray.DataArray):
        d0, d1 = sparse_array.dims
        d0_arr, d1_arr = sparse_array[d0], sparse_array[d1]
        _extent = [d1_arr[0],                  # dim1, 0
                   d1_arr[-1] + d1_arr[1],     # dim1, end + dx
                   d0_arr[-1] + d0_arr[1],     # dim0, end + dx
                   d0_arr[0]]                  # dim0, 0
    else:
        _extent = [0, sparse_array.shape[1],
                   sparse_array.shape[0], 0]

    # process the clip field
    #  if first argument is given and percentile
    if (not (alpha_clip[0] is None)) and (clip_type == 'percentile'):
        amin = np.nanpercentile(sparse_array, alpha_clip[0])
    #  if first argument is given and value
    elif (not (alpha_clip[0] is None)) and (clip_type == 'value'):
        amin = alpha_clip[0]
    #  if first argument is not given
    else:
        amin = np.nanmin(sparse_array)
    #  if second argument is given and percentile
    if (not (alpha_clip[1] is None)) and (clip_type == 'percentile'):
        amax = np.nanpercentile(sparse_array, alpha_clip[1])
    #  if second argument is given and value
    elif (not (alpha_clip[1] is None)) and (clip_type == 'value'):
        amax = alpha_clip[1]
    #  if second argument is not given
    else:
        amax = np.nanmax(sparse_array)

    # normalize the alpha channel
    alphas = matplotlib.colors.Normalize(
        amin, amax, clip=True)(sparse_array)  # Normalize alphas

    # normalize the colors
    colors = matplotlib.colors.Normalize(
        np.nanmin(sparse_array),
        np.nanmax(sparse_array))(sparse_array)  # Normalize colors

    colors = cmap(colors)
    colors[..., -1] = alphas

    im = ax.imshow(
        colors, extent=_extent)

    return im



import os
import copy
import abc
import warnings

import numpy as np
import xarray as xr

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D  # noqa: F401 unused import

from . import io
from . import plan
from . import section
from . import strat
from . import plot


class BaseCube(abc.ABC):
    """Base cube object.

    Cube objects contain t-x-y or z-x-y information.

    This base class should not be used directly, but is subclassed below,
    providing convenient obejcts for maniplating common data types.

    .. note::
        `Cube` does not load any data into memory by default. This means that
        slicing is handled "behind the scenes" by an :doc:`I/O file handler
        </reference/io/index>`. Optionally, you can load files into memory for
        (sometimes) faster operations.
        See the :meth:`read` for more information.

    """

    def __init__(self, data, read=[], varset=None, dimensions=None):
        """Initialize the BaseCube.

        Parameters
        ----------
        data : :obj:`str`, :obj:`dict`
            If data is type `str`, the string points to a NetCDF or HDF5 file
            that can be read. Typically this is used to directly import files
            output from the pyDeltaRCM model. Alternatively, pass a
            :obj:`dict` with keys indicating variable names, and values with
            corresponding t-x-y `ndarray` of data.

        read : :obj:`bool`, optional
            Which variables to read from dataset into memory. Special option
            for ``read=True`` to read all available variables into memory.

        varset : :class:`~deltametrics.plot.VariableSet`, optional
            Pass a `~deltametrics.plot.VariableSet` instance if you wish
            to style this cube similarly to another cube.

        dimensions : `dict`, optional
            A dictionary with names and coordinates for dimensions of the
            cube, if instantiating the cube from data loaded in memory
            in a dictionary.
        """
        if type(data) is str:
            # handle a path to netCDF file
            self._data_path = data
            self._connect_to_file(data_path=data)
            self._read_meta_from_file()
        elif type(data) is dict:
            # handle a dict, arrays set up already, make an io class to wrap it
            self._data_path = None
            self._dataio = io.DictionaryIO(
                data, dimensions=dimensions)
            self._read_meta_from_file()
        elif isinstance(data, DataCube):
            # handle initializing one cube type from another
            self._data_path = data.data_path
            self._dataio = data._dataio
            self._read_meta_from_file()
        else:
            raise TypeError('Invalid type for "data": %s' % type(data))

        self._planform_set = {}
        self._section_set = {}

        if varset:
            self.varset = varset
        else:
            self.varset = plot.VariableSet()

        # some undocumented aliases
        self.plans = self._planform_set
        self.plan_set = self._planform_set

    @abc.abstractmethod
    def __getitem__(self, var):
        """Return the variable.

        Overload slicing operations for io to return correct var. Must be
        implemented by subclasses.
        """
        ...

    def _connect_to_file(self, data_path):
        """Connect to file.

        This method is used internally to send the ``data_path`` to the
        correct IO handler.
        """
        _, ext = os.path.splitext(data_path)
        if ext == '.nc':
            self._dataio = io.NetCDFIO(data_path, 'netcdf')
        elif ext == '.hdf5':
            self._dataio = io.NetCDFIO(data_path, 'hdf5')
        else:
            raise ValueError(
                'Invalid file extension for "data_path": %s' % data_path)

    def _read_meta_from_file(self):
        """Read metadata information from variables in file.

        This method is used internally to gather some useful info for
        navigating the variable trees in the stored files.
        """
        self._coords = self._dataio.known_coords
        self._variables = self._dataio.known_variables

        # process the dimensions into attributes
        if len(self._dataio.dims) > 0:
            d0, d1, d2 = self._dataio.dims
        else:
            # try taking a slice of the dataset to get the coordinates
            d0, d1, d2 = self.dataio[self._dataio.known_variables[0]].dims
        self._dim0_idx = self._dataio[d0]

        # if x is 2-D then we assume x and y are mesh grid values
        if np.ndim(self._dataio[d1]) == 2:
            self._dim1_idx = self._dataio.dataset[d1][:, 0].squeeze()
            self._dim2_idx = self._dataio.dataset[d2][0, :].squeeze()
        # if x is 1-D we do mesh-gridding
        elif np.ndim(self._dataio[d1]) == 1:
            self._dim1_idx = self._dataio[d1]
            self._dim2_idx = self._dataio[d2]
        else:
            raise TypeError(
                'Shape of coordinate array was not 1d or 2d. '
                'Maybe the name was not correctly identified in the dataset, '
                'or the array is misformatted.')

        # assign values to dimensions of cube
        self._dim0_coords = self._t = self._dim0_idx
        self._dim1_coords = self._dim1_idx
        self._dim2_coords = self._dim2_idx

        # DEVELOPER NOTE: can we remvoe the _dimX_idx altogether and just use
        # the _dimX_coords arrays?

    def read(self, variables):
        """Read variable into memory.

        Parameters
        ----------
        variables : :obj:`list` of :obj:`str`, :obj:`str`
            Which variables to read into memory.
        """
        if variables is True:  # special case, read all variables
            variables = self.variables
        elif type(variables) is str:
            variables = [variables]
        else:
            raise TypeError('Invalid type for "variables": %s ' % variables)

        for var in variables:
            self._dataio.read(var)

    @property
    def meta(self):
        return self._dataio.meta

    @property
    def varset(self):
        """:class:`~deltametrics.plot.VariableSet` : Variable styling for plotting.

        Can be set with :code:`cube.varset = VariableSetInstance` where
        ``VariableSetInstance`` is a valid instance of
        :class:`~deltametrics.plot.VariableSet`.
        """
        return self._varset

    @varset.setter
    def varset(self, var):
        if type(var) is plot.VariableSet:
            self._varset = var
        else:
            raise TypeError('Pass a valid VariableSet instance.')

    @property
    def data_path(self):
        """:obj:`str` : Path connected to for file IO.

        Returns a string if connected to file, or None if cube initialized
        from ``dict``.
        """
        return self._data_path

    @property
    def dataio(self):
        """:obj:`~deltametrics.io.BaseIO` subclass : Data I/O handler.
        """
        return self._dataio

    @property
    def coords(self):
        """`list` : List of coordinate names as strings."""
        return self._coords

    @property
    def variables(self):
        """`list` : List of variable names as strings.
        """
        return self._variables

    @property
    def planform_set(self):
        """:obj:`dict` : Set of planform instances.
        """
        return self._planform_set

    @property
    def planforms(self):
        """`dict` : Set of plan instances.

        Alias to :meth:`plan_set`.
        """
        return self._planform_set

    def register_plan(self, *args, **kwargs):
        """wrapper, might not really need this."""
        return self.register_planform(*args, **kwargs)

    def register_planform(self, name, PlanformInstance, return_planform=False):
        """Register a planform to the :attr:`planform_set`.

        Connect a planform to the cube.

        Parameters
        ----------
        name : :obj:`str`
            The name to register the `Planform`.

        PlanformInstance : :obj:`~deltametrics.planform.BasePlanform` subclass instance
            The planform instance that will be registered.

        return_planform : :obj:`bool`
            Whether to return the planform object.
        """
        if not issubclass(type(PlanformInstance), plan.BasePlanform):
            raise TypeError(
                '`PlanformInstance` was not a `Planform`. '
                'Instead, was: {0}'.format(type(PlanformInstance)))
        if not isinstance(name, str):
            raise TypeError(
                '`name` was not a string. '
                'Instead, was: {0}'.format(type(name)))
        PlanformInstance.connect(self, name=name)  # attach cube
        self._planform_set[name] = PlanformInstance
        if return_planform:
            return self._planform_set[name]

    @property
    def section_set(self):
        """:obj:`dict` : Set of section instances.
        """
        return self._section_set

    @property
    def sections(self):
        """:obj:`dict` : Set of section instances.

        Alias to :meth:`section_set`.
        """
        return self._section_set

    def register_section(self, name, SectionInstance, return_section=False):
        """Register a section to the :attr:`section_set`.

        Connect a section to the cube.

        Parameters
        ----------
        name : :obj:`str`
            The name to register the section.

        SectionInstance : :obj:`~deltametrics.section.BaseSection` subclass instance
            The section instance that will be registered.

        return_section : :obj:`bool`
            Whether to return the section object.

        Notes
        -----

        When the API for instantiation of the different section types is
        settled, we should enable the ability to pass section kwargs to this
        method, and then instantiate the section internally. This avoids the
        user having to specify ``dm.section.StrikeSection(distance=2000)`` in
        the ``register_section()`` call, and instead can do something like
        ``golf.register_section('trial', trace='strike',
        distance=2000)``.
        """
        if not issubclass(type(SectionInstance), section.BaseSection):
            raise TypeError(
                '`SectionInstance` was not a `Section`. '
                'Instead, was: {0}'.format(type(SectionInstance)))
        if not isinstance(name, str):
            raise TypeError(
                '`name` was not a string. '
                'Instead, was: {0}'.format(type(name)))
        SectionInstance.connect(self, name=name)  # attach cube
        self._section_set[name] = SectionInstance
        if return_section:
            return self._section_set[name]

    @property
    def dim0_coords(self):
        """Coordinates along the first dimension of `cube`.
        """
        return self.z

    @property
    def dim1_coords(self):
        """Coordinates along the second dimension of `cube`.
        """
        return self._dim1_coords

    @property
    def dim2_coords(self):
        """Coordinates along the third dimension of `cube`.
        """
        return self._dim2_coords

    @property
    @abc.abstractmethod
    def z(self):
        """Vertical coordinate."""
        ...

    @property
    @abc.abstractmethod
    def Z(self):
        """Vertical mesh."""
        ...

    @property
    def H(self):
        """Number of elements, vertical (height) coordinate."""
        return self._H

    @property
    def L(self):
        """Number of elements, length coordinate."""
        return self._L

    @property
    def W(self):
        """Number of elements, width coordinate."""
        return self._W

    @property
    def shape(self):
        """Number of elements in data (HxLxW)."""
        return (self.H, self.L, self.W)

    @property
    def extent(self):
        """The limits of the dim1 by dim2 plane.

        Useful for plotting.
        """
        _extent = [
            self.dim2_coords[0],                         # dim1, 0
            self.dim2_coords[-1] + self.dim2_coords[1],  # dim1, end + dx
            self.dim1_coords[-1] + self.dim1_coords[1],  # dim0, end + dx
            self.dim1_coords[0]]                         # dim0, 0
        return _extent

    def export_frozen_variable(self, var, return_cube=False):
        """Export a cube with frozen values.

        Creates a `H x L x W` `ndarray` with values from variable `var` placed
        into the array. This method is particularly useful for inputs to
        operation that will repeatedly utilize the underlying data in
        computations. Access to underlying data is comparatively slow to data
        loaded in memory, because the `Cube` utilities are configured to read
        data off-disk as needed.
        """
        if return_cube:
            raise NotImplementedError
        else:
            return self[var].load()

    def quick_show(self, var, idx=-1, axis=0, **kwargs):
        """Convenient and quick way to show a slice of the cube by `idx` and `axis`.

        .. hint::

            If neither `idx` or `axis` is specified, a planform view of the
            last index is shown.

        Parameters
        ----------
        var : :obj:`str`
            Which variable to show from the underlying dataset.

        idx : :obj:`int`, optional
            Which index along the `axis` to slice data from. Default value is
            ``-1``, the last index along `axis`.

        axis : :obj:`int`, optional
            Which axis of the underlying cube `idx` is specified for. Default
            value is ``0``, the first axis of the cube.

        **kwargs
            Keyword arguments are passed
            to :meth:`~deltametrics.plan.Planform.show` if `axis` is ``0``,
            otherwise passed
            to :meth:`~deltametrics.section.BaseSection.show`.

        Examples
        --------

        .. plot::
            :include-source:

            >>> golfcube = dm.sample_data.golf()
            >>> golfstrat = dm.cube.StratigraphyCube.from_DataCube(
            ...     golfcube, dz=0.1)
            ...
            >>> fig, ax = plt.subplots(2, 1)
            >>> golfcube.quick_show('eta', ax=ax[0])  # a Planform (axis=0)
            >>> golfstrat.quick_show('eta', idx=100, axis=2, ax=ax[1])  # a DipSection
            >>> plt.show()
        """
        if axis == 0:
            # this is a planform slice
            _obj = plan.Planform(self, idx=idx)
        elif axis == 1:
            # this is a Strike section
            _obj = section.StrikeSection(self, distance_idx=idx)
        elif axis == 2:
            # this is a Dip section
            _obj = section.DipSection(self, distance_idx=idx)
        else:
            raise ValueError(
                'Invalid `axis` specified: {0}'.format(axis))

        # use the object to handle the showing
        _obj.show(var, **kwargs)

    def show_cube(self, var, style='mesh', ve=200, ax=None):
        """Show the cube in a 3D axis.

        .. important:: requires `pyvista` package for 3d visualization.

        Parameters
        ----------
        var : :obj:`str`
            Which variable to show from the underlying dataset.

        style : :obj:`str`, optional
            Style to show `cube`. Default is `'mesh'`, which gives a 3D
            volumetric view. Other supported option is `'fence'`, which gives
            a fence diagram with one slice in each cube dimension.

        ve : :obj:`float`
            Vertical exaggeration. Default is ``200``.

        ax : :obj:`~matplotlib.pyplot.Axes` object, optional
            A `matplotlib` `Axes` object to plot the section. Optional; if not
            provided, a call is made to ``plt.gca()`` to get the current (or
            create a new) `Axes` object.

        Examples
        --------

        .. note::

            The following code snippets are not set up to actually make the
            plots in the documentation.

        .. code::

            >>> golfcube = dm.sample_data.golf()
            >>> golfstrat = dm.cube.StratigraphyCube.from_DataCube(
            ...     golfcube, dz=0.1)
            ...
            >>> fig, ax = plt.subplots()
            >>> golfstrat.show_cube('eta', ax=ax)

        .. code::

            >>> golfcube = dm.sample_data.golf()
            ...
            >>> fig, ax = plt.subplots()
            >>> golfcube.show_cube('velocity', style='fence', ax=ax)
        """
        try:
            import pyvista as pv
        except ImportError:
            ImportError(
                '3d plotting dependency, pyvista, was not found.')
        except ModuleNotFoundError:
            ModuleNotFoundError(
                '3d plotting dependency, pyvista, was not found.')
        except Exception as e:
            raise e

        if not ax:
            ax = plt.gca()

        _data = np.array(self[var])
        _data = _data.transpose((2, 1, 0))

        mesh = pv.UniformGrid(_data.shape)
        mesh[var] = _data.ravel(order='F')
        mesh.spacing = (self.dim2_coords[1],
                        self.dim1_coords[1],
                        ve/self.dim1_coords[1])
        mesh.active_scalars_name = var

        p = pv.Plotter()
        p.add_mesh(mesh.outline(), color="k")
        if style == 'mesh':

            threshed = mesh.threshold([-np.inf, np.inf], all_scalars=True)
            p.add_mesh(threshed, cmap=self.varset[var].cmap)

        elif style == 'fence':
            # todo, improve this to manually create the sections so you can
            #   do more than three slices
            slices = mesh.slice_orthogonal()
            p.add_mesh(slices, cmap=self.varset[var].cmap)

        else:
            raise ValueError('Bad value for style: {0}'.format(style))

        p.show()

    def show_plan(self, *args, **kwargs):
        """Deprecated. Use :obj:`quick_show` or :obj:`show_planform`.

        .. warning::

            Provides a legacy option to quickly show a planform, from before
            the `Planform` object was properly implemented. Will be removed
            in a future release.

        Parameters
        ----------
        """
        # legacy method, ported over to show_planform.
        warnings.warn(
            '`show_plan` is a deprecated method, and has been replaced by two '
            'alternatives. To quickly show a planform slice of a cube, you '
            'can use `quick_show()` with a similar API. The `show_planform` '
            'method implements more features, but requires instantiating a '
            '`Planform` object first. Passing arguments to `quick_show`.')
        # pass `t` arg to `idx` for legacy
        if 't' in kwargs.keys():
            idx = kwargs.pop('t')
            kwargs['idx'] = idx

        self.quick_show(*args, **kwargs)

    def show_planform(self, name, variable, **kwargs):
        """Show a registered planform by name and variable.

        Call a registered planform by name and variable.

        Parameters
        ----------
        name : :obj:`str`
            The name of the registered planform.

        variable : :obj:`str`
            The varaible name to show.

        **kwargs
            Keyword arguments passed
            to :meth:`~deltametrics.plan.Planform.show`.
        """
        # call `show()` from string
        if isinstance(name, str):
            self._planform_set[name].show(variable, **kwargs)
        else:
            raise TypeError(
                '`name` was not a string, '
                'was {0}'.format(type(name)))

    def show_section(self, name, variable, **kwargs):
        """Show a registered section by name and variable.

        Call a registered section by name and variable.

        Parameters
        ----------
        name : :obj:`str`
            The name of the registered section.

        variable : :obj:`str`
            The varaible name to show.

        **kwargs
            Keyword arguments passed
            to :meth:`~deltametrics.section.BaseSection.show`.
        """
        # call `show()` from string
        if isinstance(name, str):
            self._section_set[name].show(variable, **kwargs)
        else:
            raise TypeError(
                '`name` was not a string, '
                'was {0}'.format(type(name)))


class DataCube(BaseCube):
    """DataCube object.

    DataCube contains t-x-y information. It may have any
    number of attached attributes (grain size, mud frac, elevation).
    """

    def __init__(self, data, read=[], varset=None, stratigraphy_from=None,
                 dimensions=None):
        """Initialize the BaseCube.

        Parameters
        ----------
        data : :obj:`str`, :obj:`dict`
            If data is type `str`, the string points to a NetCDF or HDF5 file
            that can be read. Typically this is used to directly import files
            output from the pyDeltaRCM model. Alternatively, pass a
            :obj:`dict` with keys indicating variable names, and values with
            corresponding t-x-y `ndarray` of data.

        read : :obj:`bool`, optional
            Which variables to read from dataset into memory. Special option
            for ``read=True`` to read all available variables into memory.

        varset : :class:`~deltametrics.plot.VariableSet`, optional
            Pass a `~deltametrics.plot.VariableSet` instance if you wish
            to style this cube similarly to another cube. If no argument is
            supplied, a new default VariableSet instance is created.

        stratigraphy_from : :obj:`str`, optional
            Pass a string that matches a variable name in the dataset to
            compute preservation and stratigraphy using that variable as
            elevation data. Typically, this is ``'eta'`` in pyDeltaRCM model
            outputs. Stratigraphy can be computed on an existing data cube
            with the :meth:`~deltametrics.cube.DataCube.stratigraphy_from`
            method.

        dimensions : `dict`, optional
            A dictionary with names and coordinates for dimensions of the
            `DataCube`, if instantiating the cube from data loaded in memory
            in a dictionary.
        """
        super().__init__(data, read, varset, dimensions=dimensions)

        # set up the grid for time
        _, self._T, _ = np.meshgrid(
            self.dim1_coords, self.dim0_coords, self.dim2_coords)

        # get shape from a variable that is not x, y, or time
        i = 0
        while i < len(self.variables):
            if self.variables[i] == 'x' or self.variables[i] == 'y' \
               or self.variables[i] == 'time':
                i += 1
            else:
                _var = self.variables[i]
                i = len(self.variables)

        # set up dimension and coordinate fields for when slices of the cube
        # are made
        self._view_dimensions = self._dataio.dims
        self._view_coordinates = copy.deepcopy({
            self._view_dimensions[0]: self.dim0_coords,
            self._view_dimensions[1]: self.dim1_coords,
            self._view_dimensions[2]: self.dim2_coords})

        # set the shape of the cube
        self._H, self._L, self._W = self[_var].data.shape

        # determine stratigraphy information
        self._knows_stratigraphy = False
        if stratigraphy_from:
            self.stratigraphy_from(variable=stratigraphy_from)

    def __getitem__(self, var):
        """Return the variable.

        Overload slicing operations for io to return a
        :obj:`~deltametrics.cube.CubeVariable` instance when slicing.

        Parameters
        ----------
        var : :obj:`str`
            Which variable to slice.

        Returns
        -------
        CubeVariable : `~deltametrics.cube.CubeVariable`
            The instantiated CubeVariable.
        """
        if var == 'time':  # special case for time
            # use the name of the first dimension, to enable
            #   unlabeled np.ndarrays and flexible name for time
            dim0_name = self.dataio.dims[0]
            dim0_coord = np.array(self.dataio.dataset[dim0_name])
            _t = np.expand_dims(dim0_coord,
                                axis=(1, 2))
            _xrt = xr.DataArray(
                np.tile(_t, (1, *self.shape[1:])),
                coords=self._view_coordinates,
                dims=self._view_dimensions)
            _obj = _xrt
        elif var in self._coords:
            # ensure coords can be called by cube[var]
            _obj = self._dataio.dataset[var]

        elif var in self._variables:
            _obj = self._dataio.dataset[var]

        else:
            raise AttributeError('No variable of {cube} named {var}'.format(
                                 cube=str(self), var=var))

        # make _obj xarray if it not already
        if isinstance(_obj, np.ndarray):
            _obj = xr.DataArray(
                _obj,
                coords=self._view_coordinates,
                dims=self._view_dimensions)
        return _obj

    def stratigraphy_from(self, variable='eta', style='mesh', **kwargs):
        """Compute stratigraphy attributes.

        Parameters
        ----------
        variable : :obj:`str`, optional
            Which variable to use as elevation data for computing
            preservation. If no value is given for this parameter, we try to
            find a variable `eta` and use that for elevation data if it
            exists.

        style : :obj:`str`, optional
            Which style of stratigraphy to compute, options are :obj:`'mesh'
            <deltametrics.strat.MeshStratigraphyAttributes>` or :obj:`'boxy'
            <deltametrics.strat.BoxyStratigraphyAttributes>`. Additional
            keyword arguments are passed to stratigraphy attribute
            initializers.

        **kwargs
            Keyword arguments passed to stratigraphy initialization. Can
            include specification for vertical resolution in `Boxy` case,
            see :obj:_determine_strat_coordinates`.
        """
        if style == 'mesh':
            self.strat_attr = \
                strat.MeshStratigraphyAttributes(elev=self[variable],
                                                 **kwargs)
        elif style == 'boxy':
            self.strat_attr = \
                strat.BoxyStratigraphyAttributes(elev=self[variable],
                                                 **kwargs)
        else:
            raise ValueError('Bad "style" argument supplied: %s' % str(style))
        self._knows_stratigraphy = True

    @property
    def z(self):
        """Vertical coordinate."""
        return self.t

    @property
    def Z(self):
        """Vertical mesh."""
        return self.T

    @property
    def t(self):
        """time coordinate."""
        return self._t

    @property
    def T(self):
        """Vertical mesh."""
        return self._T


class StratigraphyCube(BaseCube):
    """StratigraphyCube object.

    A cube of precomputed stratigraphy. This is a z-x-y matrix defining
    variables at specific voxel locations.

    This is a special case of a cube.

    """
    @staticmethod
    def from_DataCube(DataCubeInstance, stratigraphy_from='eta',
                      sigma_dist=None, dz=None, z=None, nz=None):
        """Create from a DataCube.

        Examples
        --------
        Create a stratigraphy cube from the example ``golf``:

        >>> golfcube = dm.sample_data.golf()
        >>> stratcube = dm.cube.StratigraphyCube.from_DataCube(golfcube, dz=0.05)

        Parameters
        ----------
        DataCubeInstance : :obj:`DataCube`
            The `DataCube` instance to derive from for the new
            StratigraphyCube.

        stratigraphy_from : :obj:`str`, optional
            A string that matches a variable name in the dataset to
            compute preservation and stratigraphy using that variable as
            elevation data. Typically, this is ``'eta'`` in pyDeltaRCM model
            outputs.

        **kwargs
            Keyword arguments passed to stratigraphy initialization. Can
            include specification for vertical resolution in `Boxy` case,
            see :obj:`~deltametrics.strat._determine_strat_coordinates`,
            as well as information about subsidence,
            see :obj:`~deltametrics.strat._adjust_elevation_by_subsidence`.

        Returns
        -------
        StratigraphyCubeInstance : :obj:`StratigraphyCube`
            The new `StratigraphyCube` instance.
        """
        return StratigraphyCube(DataCubeInstance,
                                varset=DataCubeInstance.varset,
                                stratigraphy_from=stratigraphy_from,
                                sigma_dist=sigma_dist, dz=dz, z=z, nz=nz)

    def __init__(self, data, read=[], varset=None,
                 stratigraphy_from=None, sigma_dist=None,
                 dz=None, z=None, nz=None):
        """Initialize the StratigraphicCube.

        Any instantiation pathway must configure :obj:`z`, :obj:`H`, :obj:`L`,
        :obj:`W`, and :obj:`strata`.

        Parameters
        ----------
        data : :obj:`str`, :obj:`dict`
            If data is type `str`, the string points to a NetCDF or HDF5 file
            that can be read. Typically this is used to directly import files
            output from the pyDeltaRCM model. Alternatively, pass a
            :obj:`dict` with keys indicating variable names, and values with
            corresponding t-x-y `ndarray` of data.

        read : :obj:`bool`, optional
            Which variables to read from dataset into memory. Special option
            for ``read=True`` to read all available variables into memory.

        varset : :class:`~deltametrics.plot.VariableSet`, optional
            Pass a `~deltametrics.plot.VariableSet` instance if you wish
            to style this cube similarly to another cube. If no argument is
            supplied, a new default VariableSet instance is created.
        """
        super().__init__(data, read, varset)
        if isinstance(data, str):
            raise NotImplementedError('Precomputed NetCDF?')
        elif isinstance(data, np.ndarray):
            raise NotImplementedError('Precomputed numpy array?')
        elif isinstance(data, DataCube):
            # i.e., creating from a DataCube
            _elev = copy.deepcopy(data[stratigraphy_from])

            # set up coordinates of the array
            if sigma_dist is not None:
                _elev_adj = strat._adjust_elevation_by_subsidence(
                    _elev.data, sigma_dist)
            else:
                _elev_adj = _elev.data
            _z = strat._determine_strat_coordinates(
                _elev_adj, dz=dz, z=z, nz=nz)
            self._z = xr.DataArray(_z, name='z', dims=['z'], coords={'z': _z})
            self._H = len(self.z)
            self._L, self._W = _elev.shape[1:]
            self._Z = np.tile(self.z, (self.W, self.L, 1)).T
            self._sigma_dist = sigma_dist

            _out = strat.compute_boxy_stratigraphy_coordinates(
                _elev_adj, sigma_dist=None, z=_z, return_strata=True)
            self.strata_coords, self.data_coords, self.strata = _out
        else:
            raise TypeError('No other input types implemented yet.')

        self._view_dimensions = ['z', *data._dataio.dims[1:]]
        self._view_coordinates = copy.deepcopy({
            self._view_dimensions[0]: self._z,
            self._view_dimensions[1]: self.dim1_coords,
            self._view_dimensions[2]: self.dim2_coords})

    def __getitem__(self, var):
        """Return the variable.

        Overload slicing operations for io to return a
        :obj:`~deltametrics.cube.CubeVariable` instance when slicing, where
        the data have been placed into stratigraphic position.

        Parameters
        ----------
        var : :obj:`str`
            Which variable to slice.

        Returns
        -------
        CubeVariable : `~deltametrics.cube.CubeVariable`
            The instantiated CubeVariable.
        """
        if var == 'time':
            # a special attribute we add, which matches eta.shape
            #   use the name of the first dimension, to enable
            #   unlabeled np.ndarrays and flexible name for time
            dim0_name = self.dataio.dims[0]
            dim0_coord = np.array(self.dataio[dim0_name])
            _t = np.expand_dims(dim0_coord,
                                axis=(1, 2))
            _arr = np.full(self.shape, np.nan)
            _var = np.tile(_t, (1, *self.shape[1:]))
        elif var in self._variables:
            _arr = np.full(self.shape, np.nan)
            # _var = np.array(self.dataio[var], copy=True)
            _var = self.dataio[var]
        else:
            raise AttributeError('No variable of {cube} named {var}'.format(
                                 cube=str(self), var=var))

        # the following lines apply the data to stratigraphy mapping
        if isinstance(_var, xr.core.dataarray.DataArray):
            _vardata = _var.data
        else:
            _vardata = _var
        _cut = _vardata[self.data_coords[:, 0], self.data_coords[:, 1],
                        self.data_coords[:, 2]]
        _arr[self.strata_coords[:, 0], self.strata_coords[:, 1],
             self.strata_coords[:, 2]] = _cut
        _obj = xr.DataArray(
            _arr,
            coords=self._view_coordinates,
            dims=self._view_dimensions)
        return _obj

    @property
    def strata(self):
        """Strata surfaces.

        Elevation of stratal surfaces, matched to times in :obj:`time`.
        """
        return self._strata

    @strata.setter
    def strata(self, var):
        self._strata = var

    @property
    def z(self):
        return self._z

    @property
    def Z(self):
        """Vertical mesh."""
        return self._Z

    @property
    def sigma_dist(self):
        """Subsidence information."""
        return self._sigma_dist



import warnings

from . import rcm8 as moved_rcm8

warnings.simplefilter("default")


def rcm8():
    """DEPRECATED rcm8 data cube location.

    The `rcm8` data cube has moved to `deltametrics.sample_data.rcm8()`.
    Access is maintained here for backwards compatability, but will be removed
    in a future release.
    """
    warnings.warn(
        "The `rcm8` data cube has moved to "
        "`deltametrics.sample_data.rcm8()`. "
        "Access is maintained here for backwards compatability, "
        "but will be removed in a future release.",
        DeprecationWarning,
    )
    return moved_rcm8()



"""Functions for channel mobility metrics.

Current available mobility metrics include:
    - Dry fraction decay from Cazanacli et al 2002
    - Planform overlap from Wickert et al 2013
    - Reworking index from Wickert et al 2013
    - Channel abandonment from Liang et al 2016

Also included are functions to fit curves to the output from the mobility
functions, allowing for decay constants and timescales to be quantified.
"""

import numpy as np
from deltametrics import mask
import xarray as xr


def check_inputs(chmap, basevalues=None, basevalues_idx=None, window=None,
                 window_idx=None, landmap=None):
    """
    Check the input variable values.

    Ensures compatibility with mobility functions. Tries to convert from some
    potential input types (xarray data arrays, deltametrics masks) if possible.

    Parameters
    ----------
    chmap : list, xarray.DataArray, numpy.ndarray
        Either a list of 2-D deltametrics.mask, xarray.DataArray, or
        numpy.ndarray objects, or a t-x-y 3-D xarray.DataArray or numpy.ndarray
        with channel mask values.

    basevalues : list, int, float, optional
        List of time values to use as the base channel map. (or single value)

    basevalues_idx : list, optional
        List of time indices to use as the base channel map. (or single value)

    window : int, float, optional
        Duration of time to use as the time lag (aka how far from the basemap
        will be analyzed).

    window_idx : int, float, optional
        Duration of time in terms of indices (# of save states) to use as the
        time lag.

    landmap : list, xarray.DataArray, numpy.ndarray, optional
        Either a list of 2-D deltametrics.mask, xarray.DataArray, or
        numpy.ndarray objects, or a t-x-y 3-D xarray.DataArray or numpy.ndarray
        with land mask values.

    """
    # handle input maps - try to convert some expected types
    in_maps = {'chmap': chmap, 'landmap': landmap}
    out_maps = {'chmap': None, 'landmap': None}
    _skip = False
    for key in in_maps.keys():
        if in_maps[key] is None:
            _skip = True
        else:
            inmap = in_maps[key]
            _skip = False
        # first expect a list of objects and coerce them into xarray.dataarrays
        if (isinstance(inmap, list)) and (_skip is False):
            # depending on type convert to xarray.dataarray
            if isinstance(inmap[0], np.ndarray) is True:
                dims = ('time', 'x', 'y')  # assumes an ultimate t-x-y shape
                if len(np.shape(inmap[0])) != 2:
                    raise ValueError('Expected list to be of 2-D ndarrays.')
                coords = {'time': np.arange(1),
                          'x': np.arange(inmap[0].shape[0]),
                          'y': np.arange(inmap[0].shape[1])}
                _converted = [xr.DataArray(
                    data=np.reshape(inmap[i],
                                    (1, inmap[i].shape[0], inmap[i].shape[1])),
                    coords=coords, dims=dims)
                              for i in range(len(inmap))]
            elif issubclass(type(inmap[0]), mask.BaseMask) is True:
                _converted = [i.mask for i in inmap]
            elif isinstance(inmap[0], xr.DataArray) is True:
                _converted = inmap
            else:
                raise TypeError('Type of objects in the input list is not '
                                'a recognized type.')
            # convert the list of xr.DataArrays into a single 3-D one
            out_maps[key] = _converted[0]  # init the final 3-D DataArray
            for j in range(1, len(_converted)):
                # stack them up along the time array into a 3-D dataarray
                out_maps[key] = xr.concat(
                    (out_maps[key], _converted[j]), dim='time').astype(float)

        elif (isinstance(inmap, np.ndarray) is True) and \
          (len(inmap.shape) == 3) and (_skip is False):
            dims = ('time', 'x', 'y')  # assumes t-x-y orientation of array
            coords = {'time': np.arange(inmap.shape[0]),
                      'x': np.arange(inmap.shape[1]),
                      'y': np.arange(inmap.shape[2])}
            out_maps[key] = \
                xr.DataArray(
                    data=inmap, coords=coords, dims=dims).astype(float)

        elif (issubclass(type(inmap), mask.BaseMask) is True) and \
          (_skip is False):
            raise TypeError(
                'Cannot input a Mask directly to mobility metrics. '
                'Use a list-of-masks instead.')

        elif (isinstance(inmap, xr.core.dataarray.DataArray) is True) and \
          (len(inmap.shape) == 3) and (_skip is False):
            out_maps[key] = inmap.astype(float)

        elif _skip is False:
            raise TypeError('Input mask data type or format not understood.')

    # can't do this binary check for a list
    # if ((chmap == 0) | (chmap == 1)).all():
    #     pass
    # else:
    #     raise ValueError('chmap was not binary')

    # check basevalues and time_window types
    if (basevalues is not None):
        try:
            baselist = list(basevalues)
            # convert to indices of the time dimension
            basevalues = [np.argmin(
                np.abs(out_maps['chmap'].time.data - i))
                for i in baselist]
        except Exception:
            raise TypeError('basevalues was not a list or list-able obj.')

    if (basevalues_idx is not None):
        try:
            basevalues_idx = list(basevalues_idx)
        except Exception:
            raise TypeError('basevalues_idx was not a list or list-able obj.')

    if (basevalues is not None) and (basevalues_idx is not None):
        raise Warning(
            'basevalues and basevalues_idx supplied, using `basevalues`.')
        base_out = basevalues
    elif (basevalues is None) and (basevalues_idx is not None):
        base_out = basevalues_idx
    elif (basevalues is not None) and (basevalues_idx is None):
        base_out = basevalues
    else:
        raise ValueError('No basevalue or basevalue_idx supplied!')

    if (window is not None) and \
      (isinstance(window, int) is False) and \
      (isinstance(window, float) is False):
        raise TypeError('Input window type was not an integer or float.')
    elif (window is not None):
        # convert to index of the time dimension
        _basetime = np.min(out_maps['chmap'].time.data)  # baseline time
        _reltime = out_maps['chmap'].time.data - _basetime  # relative time
        window = int(np.argmin(np.abs(_reltime - window)) + 1)

    if (window_idx is not None) and \
      (isinstance(window_idx, int) is False) and \
      (isinstance(window_idx, float) is False):
        raise TypeError(
            'Input window_idx type was not an integer or float.')

    if (window is not None) and (window_idx is not None):
        raise Warning(
            'window and window_idx supplied, using `window`.')
        win_out = window
    elif (window is None) and (window_idx is not None):
        win_out = window_idx
    elif (window is not None) and (window_idx is None):
        win_out = window
    else:
        raise ValueError('No window or window_idx supplied!')

    # check map shapes align
    if out_maps['landmap'] is not None:
        if np.shape(out_maps['chmap']) != np.shape(out_maps['landmap']):
            raise ValueError('Shapes of chmap and landmap do not match.')

    # check that the combined basemap + timewindow does not exceed max t-index
    Kmax = np.max(base_out) + win_out
    if Kmax > out_maps['chmap'].shape[0]:
        raise ValueError('Largest basevalue + time_window exceeds max time.')

    # collect name of the first dimenstion (should be time assuming t-x-y)
    dim0 = out_maps['chmap'].dims[0]

    # return the sanitized variables
    return out_maps['chmap'], out_maps['landmap'], base_out, win_out, dim0


def calculate_channel_decay(chmap, landmap,
                            basevalues=None, basevalues_idx=None,
                            window=None, window_idx=None):
    """
    Calculate channel decay (reduction in dry fraction).

    Uses a method similar to that in Cazanacli et al 2002 to measure the
    dry fraction of a delta over time. This requires providing an input channel
    map, an input land map, choosing a set of base maps to use, and a time
    lag/time window over which to do the analysis.

    Parameters
    ----------
    chmap : list, xarray.DataArray, numpy.ndarray
        Either a list of 2-D deltametrics.mask, xarray.DataArray, or
        numpy.ndarray objects, or a t-x-y 3-D xarray.DataArray or numpy.ndarray
        with channel mask values.

    landmap : list, xarray.DataArray, numpy.ndarray
        Either a list of 2-D deltametrics.mask, xarray.DataArray, or
        numpy.ndarray objects, or a t-x-y 3-D xarray.DataArray or numpy.ndarray
        with land mask values.

    basevalues : list, int, float, optional
        List of time values to use as the base channel map. (or single value)

    basevalues_idx : list, optional
        List of time indices to use as the base channel map. (or single value)

    window : int, float, optional
        Duration of time to use as the time lag (aka how far from the basemap
        will be analyzed).

    window_idx : int, float, optional
        Duration of time in terms of indices (# of save states) to use as the
        time lag.

    Returns
    -------
    dryfrac : ndarray
        len(basevalues) x time_window 2-D array with each row representing
        the dry fraction (aka number of pixels not visited by a channel) in
        reference to a given base value at a certain time lag (column)

    """
    # sanitize the inputs first
    chmap, landmap, basevalues, time_window, dim0 = check_inputs(
        chmap, basevalues, basevalues_idx, window, window_idx, landmap)

    # initialize dry fraction array
    dims = ('base', dim0)  # base and time-lag dimensions
    coords = {'base': np.arange(len(basevalues)),
              dim0: chmap[dim0][:time_window].values}
    dryfrac = xr.DataArray(
        data=np.zeros((len(basevalues), time_window)),
        coords=coords, dims=dims)

    # loop through basevalues
    for i in range(0, len(basevalues)):
        # pull the corresponding time index for the base value
        Nbase = basevalues[i]
        # first get the dry map (non-channelized locations)
        base_dry = np.abs(chmap[Nbase, :, :] - landmap[Nbase, :, :])

        # define base landmap
        base_map = landmap[Nbase, :, :]

        # set first dry fraction at t=0
        dryfrac[i, 0] = np.sum(base_dry) / np.sum(base_map)

        # loop through the other maps to see how the dry area declines
        for Nstep in range(1, time_window):
            # get the incremental map
            chA_step = chmap[Nbase+Nstep, :, :]
            # subtract incremental map from dry map to get the new dry fraction
            base_dry -= chA_step
            # just want binary (1 = never channlized, 0 = channel visited)
            base_dry.values[base_dry.values < 0] = 0  # no need to have negative values
            # store remaining dry fraction
            dryfrac[i, Nstep] = np.sum(base_dry) / np.sum(base_map)

    return dryfrac


def calculate_planform_overlap(chmap, landmap,
                               basevalues=None, basevalues_idx=None,
                               window=None, window_idx=None):
    """
    Calculate channel planform overlap.

    Uses a method similar to that described in Wickert et al 2013 to measure
    the loss of channel system overlap with previous channel patterns. This
    requires an input channel map, land map, as well as defining the base maps
    to use and the time window over which you want to look.

    Parameters
    ----------
    chmap : list, xarray.DataArray, numpy.ndarray
        Either a list of 2-D deltametrics.mask, xarray.DataArray, or
        numpy.ndarray objects, or a t-x-y 3-D xarray.DataArray or numpy.ndarray
        with channel mask values.

    landmap : list, xarray.DataArray, numpy.ndarray
        Either a list of 2-D deltametrics.mask, xarray.DataArray, or
        numpy.ndarray objects, or a t-x-y 3-D xarray.DataArray or numpy.ndarray
        with land mask values.

    basevalues : list, int, float, optional
        List of time values to use as the base channel map. (or single value)

    basevalues_idx : list, optional
        List of time indices to use as the base channel map. (or single value)

    window : int, float, optional
        Duration of time to use as the time lag (aka how far from the basemap
        will be analyzed).

    window_idx : int, float, optional
        Duration of time in terms of indices (# of save states) to use as the
        time lag.

    Returns
    -------
    Ophi : ndarray
        A 2-D array of the normalized overlap values, array is of shape
        len(basevalues) x time_window so each row in the array represents
        the overlap values associated with a given base value and the
        columns are associated with each time lag.

    """
    # sanitize the inputs first
    chmap, landmap, basevalues, time_window, dim0 = check_inputs(
        chmap, basevalues, basevalues_idx, window, window_idx, landmap)

    # initialize D, phi and Ophi
    dims = ('base', dim0)  # base and time-lag dimensions
    coords = {'base': np.arange(len(basevalues)),
              dim0: chmap[dim0][:time_window].values}
    D = xr.DataArray(
        data=np.zeros((len(basevalues), time_window)),
        coords=coords, dims=dims)
    phi = xr.zeros_like(D)
    Ophi = xr.zeros_like(D)

    # loop through the base maps
    for j in range(0, len(basevalues)):
        # define variables associated with the base map
        fdrybase = 1 - (np.sum(chmap[basevalues[j], :, :]) /
                        np.sum(landmap[basevalues[j], :, :]))
        fwetbase = np.sum(chmap[basevalues[j], :, :]) / \
            np.sum(landmap[basevalues[j], :, :])
        # transient maps compared over the fluvial surface present in base map
        mask_map = landmap[basevalues[j], :, :]

        # loop through the transient maps associated with this base map
        for i in range(0, time_window):
            D[j, i] = np.sum(np.abs(chmap[basevalues[j], :, :]*mask_map -
                                    chmap[basevalues[j]+i, :, :]*mask_map))
            fdrystep = 1 - (np.sum(chmap[basevalues[j]+i, :, :]*mask_map) /
                            np.sum(mask_map))
            fwetstep = np.sum(chmap[basevalues[j]+i, :, :]*mask_map) / \
                np.sum(mask_map)
            phi[j, i] = fwetbase*fdrystep + fdrybase*fwetstep
            # for Ophi use a standard area in denominator, we use base area
            Ophi[j, i] = 1 - D[j, i]/(np.sum(mask_map)*phi[j, i])
    # just return the Ophi
    return Ophi


def calculate_reworking_fraction(chmap, landmap,
                                 basevalues=None, basevalues_idx=None,
                                 window=None, window_idx=None):
    """
    Calculate the reworking fraction.

    Uses a method similar to that described in Wickert et al 2013 to measure
    the reworking of the fluvial surface with time. This requires an input
    channel map, land map, as well as defining the base maps to use and the
    time window over which you want to look.

    Parameters
    ----------
    chmap : list, xarray.DataArray, numpy.ndarray
        Either a list of 2-D deltametrics.mask, xarray.DataArray, or
        numpy.ndarray objects, or a t-x-y 3-D xarray.DataArray or numpy.ndarray
        with channel mask values.

    landmap : list, xarray.DataArray, numpy.ndarray
        Either a list of 2-D deltametrics.mask, xarray.DataArray, or
        numpy.ndarray objects, or a t-x-y 3-D xarray.DataArray or numpy.ndarray
        with land mask values.

    basevalues : list, int, float, optional
        List of time values to use as the base channel map. (or single value)

    basevalues_idx : list, optional
        List of time indices to use as the base channel map. (or single value)

    window : int, float, optional
        Duration of time to use as the time lag (aka how far from the basemap
        will be analyzed).

    window_idx : int, float, optional
        Duration of time in terms of indices (# of save states) to use as the
        time lag.

    Returns
    -------
    fr : ndarray
        A 2-D array of the reworked fraction values, array is of shape
        len(basevalues) x time_window so each row in the array represents
        the overlap values associated with a given base value and the
        columns are associated with each time lag.

    """
    # sanitize the inputs first
    chmap, landmap, basevalues, time_window, dim0 = check_inputs(
        chmap, basevalues, basevalues_idx, window, window_idx, landmap)

    # initialize unreworked pixels (Nbt) and reworked fraction (fr)
    dims = ('base', dim0)  # base and time-lag dimensions
    coords = {'base': np.arange(len(basevalues)),
              dim0: chmap[dim0][:time_window].values}
    Nbt = xr.DataArray(
        data=np.zeros((len(basevalues), time_window)),
        coords=coords, dims=dims)
    fr = xr.zeros_like(Nbt)

    # loop through the base maps
    for j in range(0, len(basevalues)):
        # define variables associated with the base map
        base = basevalues[j]
        # fluvial surface is considered to be the one present in base map
        basemask = landmap[base, :, :]
        notland = len(np.where(basemask == 0)[0])  # number of not-land pixels
        basechannels = chmap[base, :, :]
        fbase = basemask - basechannels
        fdrybase = np.sum(fbase) / np.sum(basemask)

        # initialize channelmap series through time (kb) using initial chmap
        kb = np.copy(basechannels)

        # loop through the transient maps associated with this base map
        for i in range(0, time_window):
            # if i == 0 no reworking has happened yet
            if i == 0:
                Nbt[j, i] = fdrybase * np.sum(basemask)
                fr[j, i] = 0
            else:
                # otherwise situation is different

                # transient channel map withint base fluvial surface
                tmap = chmap[base+i, :, :]*basemask

                # add to kb
                kb += tmap

                # unreworked pixels are those channels have not visited
                # get this by finding all 0s left in kb and subtracting
                # the number of non-land pixels from base fluvial surface
                unvisited = len(np.where(kb == 0)[0])
                Nbt[j, i] = unvisited - notland
                fr[j, i] = 1 - (Nbt[j, i]/(np.sum(basemask)*fdrybase))

    # just return the reworked fraction
    return fr


def calculate_channel_abandonment(chmap, basevalues=None, basevalues_idx=None,
                                  window=None, window_idx=None):
    """
    Calculate channel abandonment.

    Measure the number of channelized pixels that are no longer channelized as
    a signature of channel mobility based on method in Liang et al 2016. This
    requires providing an input channel map, and setting parameters for the
    min/max values to compare to, and the time window over which the evaluation
    can be done.

    Parameters
    ----------
    chmap : list, xarray.DataArray, numpy.ndarray
        Either a list of 2-D deltametrics.mask, xarray.DataArray, or
        numpy.ndarray objects, or a t-x-y 3-D xarray.DataArray or numpy.ndarray
        with channel mask values.

    basevalues : list, int, float, optional
        List of time values to use as the base channel map. (or single value)

    basevalues_idx : list, optional
        List of time indices to use as the base channel map. (or single value)

    window : int, float, optional
        Duration of time to use as the time lag (aka how far from the basemap
        will be analyzed).

    window_idx : int, float, optional
        Duration of time in terms of indices (# of save states) to use as the
        time lag.

    Returns
    -------
    PwetA : ndarray
        A 2-D array of the abandoned fraction of the channel over the window of
        time. It is of shape len(basevalues) x time_window so each row
        represents the fraction of the channel that has been abandonded, and
        the columns are associated with each time lag.

    """
    # sanitize the inputs first
    chmap, landmap, basevalues, time_window, dim0 = check_inputs(
        chmap, basevalues, basevalues_idx, window, window_idx)
    # initialize values
    dims = ('base', dim0)  # base and time-lag dimensions
    coords = {'base': np.arange(len(basevalues)),
              dim0: chmap[dim0][:time_window].values}
    PwetA = xr.DataArray(
        data=np.zeros((len(basevalues), time_window)),
        coords=coords, dims=dims)
    chA_base = xr.zeros_like(chmap[0, :, :])
    chA_step = xr.zeros_like(chmap[0, :, :])

    # loop through the basevalues
    for i in range(0, len(basevalues)):
        # first get the 'base' channel map that is being compared to
        Nbase = basevalues[i]
        chA_base = chmap[Nbase, :, :]
        # get total number of channel pixels in that map
        baseA = np.sum(chA_base)
        # loop through the other maps to be compared against the base map
        for Nstep in range(1, time_window):
            # get the incremental map
            chA_step = chmap[Nbase+Nstep, :, :]
            # get the number of pixels that were abandonded
            stepA = len(
                np.where(chA_base.values.flatten() >
                         chA_step.values.flatten())[0])
            # store this number in the PwetA array for each transient map
            PwetA[i, Nstep] = stepA/baseA

    return PwetA


def channel_presence(chmap):
    """
    Calculate the normalized channel presence at each pixel location.

    Measure the normalized fraction of time a given pixel is channelized,
    based on method in Liang et al 2016. This requires providing a 3-D input
    channel map (t-x-y).

    Parameters
    ----------
    chmap : ndarray
        A t-x-y shaped array with binary channel maps

    Returns
    -------
    channel_presence : ndarray
        A x-y shaped array with the normalized channel presence values.

    Examples
    --------

    .. plot::
        :include-source:

        >>> golfcube = dm.sample_data.golf()
        >>> (x, y) = np.shape(golfcube['eta'][-1, ...])
        >>> # calculate channel masks/presence over final 5 timesteps
        >>> chmap = np.zeros((5, x, y))  # initialize channel map
        >>> for i in np.arange(-5, 0):
        ...     chmap[i, ...] = dm.mask.ChannelMask(
        ...         golfcube['eta'][i, ...], golfcube['velocity'][i, ...],
        ...         elevation_threshold=0, flow_threshold=0).mask
        >>>
        >>> fig, ax = plt.subplots(1, 2)
        >>> golfcube.quick_show('eta', ax=ax[0])  # final delta
        >>> p = ax[1].imshow(dm.mobility.channel_presence(chmap), cmap='Blues')
        >>> dm.plot.append_colorbar(p, ax[1], label='Channelized Time')
        >>> plt.show()

    """
    tmp_chans = None  # instantiate
    if isinstance(chmap, mask.ChannelMask) is True:
        chans = chmap._mask
    elif isinstance(chmap, np.ndarray) is True:
        tmp_chans = chmap
    elif isinstance(chmap, xr.DataArray) is True:
        chans = chmap
    elif isinstance(chmap, list) is True:
        # convert to numpy.ndarray if possible
        if (isinstance(chmap[0], np.ndarray) is True) \
          or (isinstance(chmap[0], xr.DataArray) is True):
            # init empty array
            tmp_chans = np.zeros(
                (len(chmap), chmap[0].squeeze().shape[0],
                 chmap[0].squeeze().shape[1]))
            # populate it
            for i in range(len(chmap)):
                if isinstance(chmap[0], xr.DataArray) is True:
                    tmp_chans[i, ...] = chmap[i].data.squeeze()
                else:
                    tmp_chans[i, ...] = chmap[i].squeeze()
        elif issubclass(type(chmap[0]), mask.BaseMask) is True:
            tmp_chans = [i.mask for i in chmap]
            # convert list to ndarray
            chans = np.zeros(
                (len(tmp_chans), tmp_chans[0].shape[1], tmp_chans[0].shape[2]))
            for i in range(chans.shape[0]):
                chans[i, ...] = tmp_chans[i]
        else:
            raise ValueError('Invalid values in the supplied list.')
    else:
        raise TypeError('chmap data type not understood.')
    # if tmp_chans is a numpy.ndarray, dimensions are not known
    if isinstance(tmp_chans, np.ndarray):
        dims = ('time', 'x', 'y')  # assumes an ultimate t-x-y shape
        coords = {'time': np.arange(tmp_chans.shape[0]),
                  'x': np.arange(tmp_chans.shape[1]),
                  'y': np.arange(tmp_chans.shape[2])}
        chans = xr.DataArray(data=tmp_chans, coords=coords, dims=dims)
    # calculation of channel presence is actually very simple
    channel_presence = np.sum(chans, axis=0) / chans.shape[0]
    return channel_presence



from .sample_data import (
    _get_golf_path,
    _get_xslope_path,
    _get_aeolian_path,
    _get_rcm8_path,
    _get_landsat_path,
    golf,
    xslope,
    aeolian,
    rcm8,
    landsat,
)
from . import cube




import abc
import os
import copy
from warnings import warn

import xarray as xr
import numpy as np
import netCDF4


class BaseIO(abc.ABC):
    """BaseIO object other file format wrappers inheririt from.

    .. note::

        This is an abstract class and cannot be instantiated directly. If you
        wish to subclass to create a new IO format, you must implement
        several methods.

        To create an IO format for data already loaded into memory, you can
        subclass `BaseIO` directly, and you just need to implement the
        `__getitem__` method and `keys` attribute.

        To create an IO format that reads data from disk, you should subclass
        `FileIO`, and implement the required methods `__getitem__`, `connect`,
        `read`, and `write`, and the  `keys` attribute.
    """

    def __init__(self, io_type):
        """Initialize the base IO.
        """
        self.io_type = io_type

    @abc.abstractmethod
    def __getitem__(self):
        """Should slice the data from file.
        """
        return

    @property
    @abc.abstractmethod
    def keys(self):
        """Should link to all key _names_ stored in file.
        """
        return


class FileIO(BaseIO):
    """Base class for File input output datasets.

    This class should be the basis for subclasses that read data directly from
    a file or folder.

    To create an IO format that reads data from disk, you should subclass
    `FileIO`, and implement the required methods `__getitem__`, `connect`,
    `read`, and `write`, and the  `keys` attribute.
    """

    def __init__(self, data_path, io_type, write=False):
        """Initialize a file IO handler.

        Initialize a connection to a NetCDF file.

        Parameters
        ----------
        data_path : `str`
            Path to file to read or write to.

        type : `str`
            Stores the type of output file loaded, either a netCDF4 file,
            'netcdf' or an HDF5 file, 'hdf5'.

        write : `bool`, optional
            Whether to allow writing to an existing file. Set to False by
            default, if a file already exists at ``data_path``, writing is
            disabled, unless ``write`` is set to True.
        """
        self.data_path = data_path
        self.io_type = io_type
        self.write = write

        self.connect()

        self.get_known_coords()
        self.get_known_variables()

        super().__init__(io_type=io_type)

    @property
    def data_path(self):
        """`str` : Path to data file.

        Parameters
        ----------
        data_path : str
            path to data file for IO operations.

        Notes
        -----
        The setter method validates the path, and returns a
        ``FileNotFoundError`` if the file is not found.
        """
        return self._data_path

    @data_path.setter
    def data_path(self, var):
        if os.path.exists(var):
            self._data_path = var
        else:
            raise FileNotFoundError('File not found at supplied path: %s' % var)

    @abc.abstractmethod
    def connect(self):
        """Should connect to the data file.

        This function should initialize the file if it does not exist, or
        connect to the file if it already exists---but *do not* read the file.

        If no file is required, this function should simply pass.
        """
        return

    @abc.abstractmethod
    def get_known_variables(self):
        """Should create list of known variables.

        This function needs to populate `self.known_variables`.
        """
        return

    @abc.abstractmethod
    def get_known_coords(self):
        """A list of known coordinates.

        This function needs to populate `self.known_coords`.
        """
        return

    @abc.abstractmethod
    def read(self):
        """Should read data into memory.
        """
        return

    @abc.abstractmethod
    def write(self):
        """Should write the data to file.

        Take a :obj:`~deltametrics.cube.Cube` and write it to file.
        """
        return


class NetCDFIO(FileIO):
    """Utility for consistent IO with netCDF4 files.

    This module wraps calls to the netCDF4 python module in a consistent API,
    so the user can work seamlessly with either netCDF4 files or HDF5 files.
    The public methods of this class are consistent with
    :obj:`~deltametrics.utils.HDFIO`.

    Note that the netCDF4, netCDF4-classic, and HDF5 file standards are very
    similar and (almost) interchangable. This means that the same data loader
    can be used to handle these files. We use the `xarray` data reader which
    supports the netCDF4/HDF5 file-format.

    Older file formats such as netCDF3 or HDF4 are unsupported. For more
    information about the netCDF4 format, visit the netCDF4
    `docs <https://www.unidata.ucar.edu/software/netcdf/docs/faq.html>`_.
    """

    def __init__(self, data_path, io_type, write=False):
        """Initialize the NetCDFIO handler.

        Initialize a connection to a NetCDF file.

        Parameters
        ----------
        data_path : `str`
            Path to file to read or write to.

        io_type : `str`
            Stores the type of output file loaded, either a netCDF4 file,
            'netcdf' or an HDF5 file, 'hdf5'.

        write : `bool`, optional
            Whether to allow writing to an existing file. Set to False by
            default, if a file already exists at ``data_path``, writing is
            disabled, unless ``write`` is set to True.
        """

        super().__init__(data_path=data_path, io_type=io_type, write=write)

        self._in_memory_data = {}

    def connect(self):
        """Connect to the data file.

        Initialize the file if it does not exist, or simply ``return`` if the
        file already exists. This connection to the data file is "lazy"
        loading, meaning that array values are not being loaded into memory.

        .. note::
            This function is automatically called during initialization of any
            IO object, so it is not necessary to call it directly.

        """
        if not os.path.isfile(self.data_path):
            _tempdataset = netCDF4.Dataset(
                self.data_path, "w", format="NETCDF4")
            _tempdataset.close()

        _ext = os.path.splitext(self.data_path)[-1]
        if _ext == '.nc':
            _engine = 'netcdf4'
        elif _ext == '.hdf5':
            _engine = 'h5netcdf'
        else:
            TypeError('File format is not supported '
                      'by DeltaMetrics: {0}'.format(_ext))

        try:
            # open the dataset
            _dataset = xr.open_dataset(self.data_path, engine=_engine)
        except Exception as e:
            raise TypeError(
                f'File format out of scope for DeltaMetrics: {e}')

        # try to find if coordinates have been preconfigured
        _coords_list = list(_dataset.coords)
        if len(_coords_list) == 3:
            # the coordinates are preconfigured
            self.dataset = _dataset.set_coords(_coords_list)
            self.coords = list(self.dataset.coords)
            self.dims = copy.deepcopy(self.coords)
        elif set(['total_time', 'length', 'width']).issubset(set(_dataset.dims.keys())):
            # the coordinates are not set, but there are matching arrays
            # this is a legacy option, so issue a warning here
            self.dataset = _dataset.set_coords(['x', 'y', 'time'])
            self.dims = ['time', 'length', 'width']
            self.coords = ['total_time', 'x', 'y']
            warn('Coordinates for "time", and ("y", "x") were found as '
                 'variables in the underlying data file, '
                 'but are not specified as coordinates in the undelying '
                 'data file. Please reformat the data file for use '
                 'with DeltaMetrics. This warning may be replaced '
                 'with an Error in a future version.', UserWarning)
        else:
            # coordinates were not found and are not being set
            raise NotImplementedError(
                'Underlying NetCDF datasets without any specified coordinates '
                'are not supported. See source for additional notes about '
                'how to implement this feature.')
            # DEVELOPER NOTE: it may be possible to support a netcdf file that
            # does not have specified coordinates, but we need a test case to
            # make it work. It may work to just pass everything along to the
            # cube, and let xarray automatically handle the naming of
            # coordinates, but I have not tested this.

            # self.dataset = _dataset.set_coords([])
            # self.dims = []
            # warn('Coordinates for "time", and set("x", "y") not provided in the \
            #       given data file.', UserWarning)

        try:
            _meta = xr.open_dataset(self.data_path, group='meta',
                                    engine=_engine)
            self.meta = _meta
        except OSError:
            warn('No associated metadata was found in the given data file.',
                 UserWarning)
            self.meta = None

    def get_known_variables(self):
        """List known variables.

        These variables are pulled from the loaded dataset.
        """
        _vars = list(self.dataset.variables)
        _coords = list(self.dataset.coords)
        if ('strata_age' in _vars) or ('strata_depth' in _vars):
            _coords += ['strata_age', 'strata_depth']
        self.known_variables = [item for item in _vars if item not in _coords]

    def get_known_coords(self):
        """List known coordinates.

        These coordinates are pulled from the loaded dataset.
        """
        self.known_coords = list(self.dataset.coords)

    def read(self, var):
        """Read variable from file and into memory.

        Converts `variables` in data file to `xarray` objects for coersion
        into a :obj:`~deltametrics.cube.Cube` instance.

        Parameters
        ----------
        var : `str`
            Which variable to load from the file.
        """
        try:
            _arr = self.dataset[var]
        except KeyError as e:
            raise e

        self._in_memory_data[var] = _arr.load()

    def write(self):
        """Write data to file.

        Take a :obj:`~deltametrics.cube.Cube` and write it to file.

        .. warning::
            Not Implemented.

        """
        raise NotImplementedError

    def __getitem__(self, var):
        if var in self._in_memory_data.keys():
            return self._in_memory_data[var]
        else:
            return self.dataset[var]

    @property
    def keys(self):
        """Variable names in file.
        """
        return [var for var in self.dataset.variables]


class DictionaryIO(BaseIO):
    """Utility for consistent IO with a dictionary as input.

    This module wraps calls to an underyling data dictionary, so that any
    arbitrary data can be used as a cube dataset.
    """

    def __init__(self, data_dictionary, dimensions=None):
        """Initialize the dictionary handler.

        Parameters
        ----------
        data_dictionary : `dict`
            The dictionary, containing `np.ndarray` or `xr.DataArray` data for
            each variable.
        """

        super().__init__(io_type='dictionary')

        self.dataset = data_dictionary
        self._in_memory_data = self.dataset

        self.get_known_variables()
        self.get_known_coords(dimensions)

    def get_known_variables(self):
        """List known variables.

        These variables are pulled from the loaded dataset.
        """
        _vars = self.dataset.keys()
        self.known_variables = [item for item in _vars]

    def get_known_coords(self, dimensions):
        """List known coordinates.

        These coordinates must be supplied during instantiation.
        """
        # start with a grab of underlying data
        under = list(self.dataset.values())[0]
        under_shp = under.shape

        # if the underlying data variables are xarray,
        #   then we ignore any of the other argument passed
        if isinstance(under, xr.core.dataarray.DataArray):
            # get the coordinates and dimensions from the data
            self.dims = under.dims
            self.coords = [under.coords[dim].data for dim in self.dims]
            self.dimensions = dict(zip(self.dims, self.coords))
        # otherwise, check for the arguments passed
        elif not (dimensions is None):
            # if dimensions was passed, it must be a dictionary
            if not isinstance(dimensions, dict):
                raise TypeError(
                    'Input type for `dimensions` must be '
                    '`dict` but was {0}'.format(type(dimensions)))
            # there should be exactly 3 keys
            if not (len(dimensions.keys()) == 3):
                raise ValueError(
                    '`dimensions` must contain three dimensions!')
            # use the dimensions keys as dims and the vals as coords
            #   note, we check the size against the underlying a we go
            for i, (k, v) in enumerate(dimensions.items()):
                if not (len(dimensions[k]) == under_shp[i]):
                    raise ValueError(
                        'Shape of `dimensions` at position {0} was {1}, '
                        'which does not match the variables dimensions '
                        '{2}.'.format(i, len(dimensions[k]), under_shp))
            # make the assignment
            self.dims = list(dimensions.keys())
            self.coords = list(dimensions.values())
            self.dimensions = dimensions
        # otherwise, fill with np.arange(shape)
        else:
            self.dims = ['dim0', 'dim1', 'dim2']
            coords = []
            for i in range(3):
                coords.append(np.arange(under_shp[i]))
            self.coords = coords

        self.known_coords = self.dims
        self.dimensions = dict(zip(self.dims, self.coords))

    def connect(self, *args, **kwargs):
        """Connect to the data file.

        .. warning::
            Not Implemented.

        """
        raise NotImplementedError

    def read(self, *args, **kwargs):
        """Read variable from file and into memory.

        .. warning::
            Not Implemented. Data is always in memory.

        """
        raise NotImplementedError

    def write(self):
        """Write data to file.

        Take a :obj:`~deltametrics.cube.Cube` and write it to file.

        .. warning::
            Not Implemented.

        """
        raise NotImplementedError

    def __getitem__(self, var):
        """Get item reimplemented for dictionaires.

        Returns the variables exactly as they are: either a numpy ndarray or
        xarray.
        """
        if var in self.dataset.keys():
            return self.dataset[var]
        elif var in self.known_coords:
            return self.dimensions[var]
        else:
            raise ValueError(
                'No variable named {0} found.'.format(var))

    @property
    def keys(self):
        """Variable names in file.
        """
        return [var for var in self.dataset.variables]
